"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

from multiprocessing.pool import ThreadPool
from typing import TYPE_CHECKING, Any
from urllib import parse
from urllib.parse import quote

from .config import PikaurConfig
from .core import DataType
from .exceptions import AURError
from .progressbar import ThreadSafeProgressBar
from .urllib import get_gzip_from_url, get_json_from_url

if TYPE_CHECKING:
    from .srcinfo import SrcInfo


MAX_URL_LENGTH = 8177  # default value in many web servers


class AurBaseUrl:
    aur_base_url: str | None = None

    @classmethod
    def get(cls) -> str:
        if not cls.aur_base_url:
            cls.aur_base_url = PikaurConfig().network.AurUrl.get_str()
        return cls.aur_base_url


class AURPackageInfo(DataType):
    packagebase: str
    name: str
    version: str
    desc: str | None = None
    numvotes: int | None = None
    popularity: float | None = None
    depends: list[str] = []
    makedepends: list[str] = []
    optdepends: list[str] = []
    checkdepends: list[str] = []
    conflicts: list[str] = []
    replaces: list[str] = []
    provides: list[str] = []

    aur_id: str | None = None
    packagebaseid: str | None = None
    url: str | None = None
    outofdate: int | None = None
    maintainer: str | None = None
    firstsubmitted: int | None = None
    lastmodified: int | None = None
    urlpath: str | None = None
    pkg_license: str | None = None
    keywords: list[str] = []
    groups: list[str] = []
    submitter: str | None = None
    comaintainers: list[str] = []

    @property
    def git_url(self) -> str:
        return f"{AurBaseUrl.get()}/{self.packagebase}.git"

    def __init__(self, **kwargs: Any) -> None:
        for aur_api_name, pikaur_class_name in (
            ("description", "desc", ),
            ("id", "aur_id", ),
            ("license", "pkg_license", ),
        ):
            if aur_api_name in kwargs:
                kwargs[pikaur_class_name] = kwargs.pop(aur_api_name)
        super().__init__(**kwargs)

    @classmethod
    def from_srcinfo(cls, srcinfo: "SrcInfo") -> "AURPackageInfo":
        return cls(
            name=srcinfo.package_name,
            version=(srcinfo.get_value("pkgver") or "") + "-" + (srcinfo.get_value("pkgrel") or ""),
            desc=srcinfo.get_value("pkgdesc"),
            packagebase=srcinfo.get_value("pkgbase"),
            depends=[dep.line for dep in srcinfo.get_depends().values()],
            makedepends=[dep.line for dep in srcinfo.get_build_makedepends().values()],
            checkdepends=[dep.line for dep in srcinfo.get_build_checkdepends().values()],
            **{
                key: srcinfo.get_values(key)
                for key in [
                    "optdepends",
                    "conflicts",
                    "replaces",
                    "provides",
                ]
            }
        )

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__} "{self.name}" '
            f"{self.version}>"
        )


def construct_aur_rpc_url_from_uri(uri: str) -> str:
    return AurBaseUrl.get() + "/rpc/?" + uri


def construct_aur_rpc_url_from_params(params: dict[str, str | int]) -> str:
    return construct_aur_rpc_url_from_uri(parse.urlencode(params))


def strip_aur_repo_name(pkg_name: str) -> str:
    if pkg_name.startswith("aur/"):
        return "".join(pkg_name.split("aur/"))
    return pkg_name


def aur_rpc_search_name_desc(search_query: str) -> list[AURPackageInfo]:
    url = construct_aur_rpc_url_from_params({
        "v": 5,
        "type": "search",
        "arg": strip_aur_repo_name(search_query),
        "by": "name-desc"
    })
    result_json = get_json_from_url(url)
    if "error" in result_json:
        raise AURError(url=url, error=result_json["error"])
    return [
        AURPackageInfo(
            **{key.lower(): value for key, value in aur_json.items()},
            ignore_extra_properties=True
        )
        for aur_json in result_json.get("results", [])
    ]


def _get_aur_rpc_info_url(search_queries: list[str]) -> str:
    uri = parse.urlencode({
        "v": 5,
        "type": "info",
    })
    for package in search_queries:
        uri += "&arg[]=" + quote(strip_aur_repo_name(package))
    return construct_aur_rpc_url_from_uri(uri)


def aur_rpc_info(search_queries: list[str]) -> list[AURPackageInfo]:
    url = _get_aur_rpc_info_url(search_queries=search_queries)
    result_json = get_json_from_url(url)
    if "error" in result_json:
        raise AURError(url=url, error=result_json["error"])
    return [
        AURPackageInfo(
            **{key.lower(): value for key, value in aur_json.items()},
            ignore_extra_properties=True
        )
        for aur_json in result_json.get("results", [])
    ]


def aur_rpc_info_with_progress(
        search_queries: list[str], *, progressbar_length: int, with_progressbar: bool
) -> list[AURPackageInfo]:
    result = aur_rpc_info(search_queries)
    if with_progressbar:
        progressbar = ThreadSafeProgressBar.get(
            progressbar_length=progressbar_length,
            progressbar_id="aur_search",
        )
        progressbar.update()
    return result


class AurPackageListCache:

    cache: list[str] = []

    @classmethod
    def get(cls) -> list[str]:
        if not cls.cache:
            cls.cache = get_gzip_from_url(AurBaseUrl.get() + "/packages.gz").splitlines()[1:]
        return cls.cache


class AurPackageSearchCache:

    cache: dict[str, AURPackageInfo] = {}

    @classmethod
    def put(cls, pkg: AURPackageInfo) -> None:
        cls.cache[pkg.name] = pkg

    @classmethod
    def get(cls, pkg_name: str) -> AURPackageInfo | None:
        return cls.cache.get(pkg_name)


def get_all_aur_names() -> list[str]:
    return AurPackageListCache.get()


def get_max_pkgs_chunks(package_names: list[str]) -> list[list[str]]:
    chunks = []
    chunk: list[str] = []
    pkgs_to_do = package_names[::]
    while pkgs_to_do:
        if len(_get_aur_rpc_info_url(chunk + [pkgs_to_do[0]])) < MAX_URL_LENGTH:
            chunk.append(pkgs_to_do.pop(0))
        else:
            chunks.append(chunk)
            chunk = []
    if chunk:
        chunks.append(chunk)
    return chunks


def find_aur_packages(
        package_names: list[str], *, with_progressbar: bool = False
) -> tuple[list[AURPackageInfo], list[str]]:

    # @TODO: return only packages for the current architecture
    package_names = [strip_aur_repo_name(name) for name in package_names]
    num_packages = len(package_names)
    json_results = []
    for package_name in package_names[:]:
        aur_pkg = AurPackageSearchCache.get(package_name)
        if aur_pkg:
            json_results.append(aur_pkg)
            package_names.remove(package_name)

    if package_names:
        with ThreadPool() as pool:
            search_chunks = get_max_pkgs_chunks(package_names)
            requests = [
                pool.apply_async(aur_rpc_info_with_progress, [], {
                    "search_queries": chunk,
                    "progressbar_length": len(search_chunks),
                    "with_progressbar": with_progressbar
                })
                for chunk in search_chunks
            ]
            pool.close()
            results = [request.get() for request in requests]
            pool.join()
            for result in results:
                for aur_pkg in result:
                    AurPackageSearchCache.put(aur_pkg)
                    if aur_pkg.name in package_names:
                        json_results.append(aur_pkg)

    found_aur_packages = [
        result.name for result in json_results
    ]
    not_found_packages: list[str] = (
        [] if num_packages == len(found_aur_packages)
        else [
            package for package in package_names
            if package not in found_aur_packages
        ]
    )
    return json_results, not_found_packages


def get_repo_url(package_base_name: str) -> str:
    return f"{AurBaseUrl.get()}/{package_base_name}.git"


def get_all_aur_packages() -> list[AURPackageInfo]:
    return find_aur_packages(get_all_aur_names(), with_progressbar=True)[0]



# Forked from python's stdlib argparse.ArgumentParser
# Original author: Steven J. Bethard <steven.bethard@gmail.com>.
# pylint: disable=too-many-statements,too-many-locals,too-many-branches,protected-access

from argparse import SUPPRESS, Action, ArgumentError, ArgumentParser, Namespace, _get_action_name

from .i18n import translate as _


class ArgumentParserWithUnknowns(ArgumentParser):

    def _parse_known_args(
            self, arg_strings: list[str], namespace: Namespace
    ) -> tuple[Namespace, list[str]]:
        # replace arg strings that are file references
        if self.fromfile_prefix_chars is not None:
            arg_strings = self._read_args_from_files(arg_strings)

        # map all mutually exclusive arguments to the other arguments
        # they can't occur with
        action_conflicts: dict[Action, list[Action]] = {}
        for mutex_group in self._mutually_exclusive_groups:
            group_actions = mutex_group._group_actions
            for i, mutex_action in enumerate(mutex_group._group_actions):
                conflicts = action_conflicts.setdefault(mutex_action, [])
                conflicts.extend(group_actions[:i])
                conflicts.extend(group_actions[i + 1:])

        # find all option indices, and determine the arg_string_pattern
        # which has an 'O' if there is an option at an index,
        # an 'A' if there is an argument, or a '-' if there is a '--'
        option_string_indices = {}
        arg_string_pattern_parts = []
        arg_strings_iter = iter(arg_strings)
        for i, arg_string in enumerate(arg_strings_iter):

            # all args after -- are non-options
            if arg_string == "--":
                arg_string_pattern_parts.append("-")
                for _arg_string in arg_strings_iter:
                    arg_string_pattern_parts.append("A")

            # otherwise, add the arg to the arg strings
            # and note the index if it was an option
            else:
                option_tuple = self._parse_optional(arg_string)
                if option_tuple is None:
                    pattern = "A"
                else:
                    option_string_indices[i] = option_tuple
                    pattern = "O"
                arg_string_pattern_parts.append(pattern)

        # join the pieces together to form the pattern
        arg_strings_pattern = "".join(arg_string_pattern_parts)

        # converts arg strings to the appropriate and then takes the action
        seen_actions = set()
        seen_non_default_actions = set()

        def take_action(
                action: Action, argument_strings: list[str], option_string: str | None = None
        ) -> None:
            seen_actions.add(action)
            argument_values = self._get_values(action, argument_strings)

            # error if this argument is not allowed with other previously
            # seen arguments, assuming that actions that use the default
            # value don't really count as "present"
            if argument_values is not action.default:
                seen_non_default_actions.add(action)
                for conflict_action in action_conflicts.get(action, []):
                    if conflict_action in seen_non_default_actions:
                        msg = _("not allowed with argument %s")
                        action_name = _get_action_name(conflict_action)
                        raise ArgumentError(action, msg % action_name)

            # take the action if we didn't receive a SUPPRESS value
            # (e.g. from a default)
            if argument_values is not SUPPRESS:
                action(self, namespace, argument_values, option_string)

        # function to convert arg_strings into an optional action
        def consume_optional(start_index: int) -> tuple[int, list[str]]:

            unknown_args: list[str] = []

            # get the optional identified at this index
            option_tuple = option_string_indices[start_index]
            action, option_string, explicit_arg = option_tuple

            # identify additional optionals in the same arg string
            # (e.g. -xyz is the same as -x -y -z if no args are required)
            match_argument = self._match_argument
            action_tuples: list[tuple[Action, list[str], str]] = []
            while True:

                # if we found no optional action, skip it
                if action is None:
                    extras.append(arg_strings[start_index])
                    return start_index + 1, unknown_args

                # if there is an explicit argument, try to match the
                # optional's string arguments to only this
                if explicit_arg is not None:
                    arg_count = match_argument(action, "A")

                    # if the action is a single-dash option and takes no
                    # arguments, try to parse more single-dash options out
                    # of the tail of the option string
                    chars = self.prefix_chars
                    if arg_count == 0 and option_string[1] not in chars:
                        action_tuples.append((action, [], option_string))
                        char = option_string[0]
                        option_string = char + explicit_arg[0]
                        new_explicit_arg = explicit_arg[1:] or None
                        optionals_map = self._option_string_actions
                        if option_string in optionals_map:
                            action = optionals_map[option_string]
                            explicit_arg = new_explicit_arg
                        else:
                            # if we encountered unknown arg
                            # return it and add later to other
                            # unknown args
                            unknown_args.append(option_string)
                            start_index += 1
                            explicit_arg = "".join(explicit_arg[1:])
                            if explicit_arg == "":
                                stop = start_index
                                break

                    # if the action expect exactly one argument, we've
                    # successfully matched the option; exit the loop
                    elif arg_count == 1:  # noqa: RET508
                        stop = start_index + 1
                        args = [explicit_arg]
                        action_tuples.append((action, args, option_string))
                        break

                    # error if a double-dash option did not use the
                    # explicit argument
                    else:
                        msg = _("ignored explicit argument %r")
                        raise ArgumentError(action, msg % explicit_arg)

                # if there is no explicit argument, try to match the
                # optional's string arguments with the following strings
                # if successful, exit the loop
                else:
                    start = start_index + 1
                    selected_patterns = arg_strings_pattern[start:]
                    arg_count = match_argument(action, selected_patterns)
                    stop = start + arg_count
                    args = arg_strings[start:stop]
                    action_tuples.append((action, args, option_string))
                    break

            # add the Optional to the list and return the index at which
            # the Optional's string args stopped
            if not action_tuples:
                raise RuntimeError()
            for action, args, option_string in action_tuples:
                take_action(action, args, option_string)
            return stop, unknown_args

        # the list of Positionals left to be parsed; this is modified
        # by consume_positionals()
        positionals = self._get_positional_actions()

        # function to convert arg_strings into positional actions
        def consume_positionals(start_index: int) -> int:
            # match as many Positionals as possible
            match_partial = self._match_arguments_partial
            selected_pattern = arg_strings_pattern[start_index:]
            arg_counts = match_partial(positionals, selected_pattern)

            # slice off the appropriate arg strings for each Positional
            # and add the Positional and its args to the list
            for action, arg_count in zip(positionals, arg_counts, strict=True):
                args = arg_strings[start_index: start_index + arg_count]
                start_index += arg_count
                take_action(action, args)

            # slice off the Positionals that we just parsed and return the
            # index at which the Positionals' string args stopped
            positionals[:] = positionals[len(arg_counts):]
            return start_index

        # consume Positionals and Optionals alternately, until we have
        # passed the last option string
        extras = []
        start_index = 0
        if option_string_indices:
            max_option_string_index = max(option_string_indices)
        else:
            max_option_string_index = -1
        while start_index <= max_option_string_index:

            # consume any Positionals preceding the next option
            next_option_string_index = min(
                index
                for index in option_string_indices
                if index >= start_index)
            if start_index != next_option_string_index:
                positionals_end_index = consume_positionals(start_index)

                # only try to parse the next optional if we didn't consume
                # the option string during the positionals parsing
                if positionals_end_index > start_index:
                    start_index = positionals_end_index
                    continue
                start_index = positionals_end_index

            # if we consumed all the positionals we could and we're not
            # at the index of an option string, there were extra arguments
            if start_index not in option_string_indices:
                strings = arg_strings[start_index:next_option_string_index]
                extras.extend(strings)
                start_index = next_option_string_index

            # consume the next optional and any arguments for it
            start_index, unknown_args = consume_optional(start_index)
            extras.extend(unknown_args)

        # consume any positionals following the last Optional
        stop_index = consume_positionals(start_index)

        # if we didn't consume all the argument strings, there were extras
        extras.extend(arg_strings[stop_index:])

        # make sure all required actions were present and also convert
        # action defaults which were not given as arguments
        required_actions: list[str] = []
        for action in self._actions:
            if action not in seen_actions:
                action_name = _get_action_name(action)
                if action.required and action_name:
                    required_actions.append(action_name)
                else:
                    # Convert action default now instead of doing it before
                    # parsing arguments to avoid calling convert functions
                    # twice (which may fail) if the argument was given, but
                    # only if it was defined already in the namespace
                    if (
                            action.default is not None and
                            isinstance(action.default, str) and
                            hasattr(namespace, action.dest) and
                            action.default is getattr(namespace, action.dest)
                    ):
                        setattr(namespace, action.dest,
                                self._get_value(action, action.default))

        if required_actions:
            self.error(_("the following arguments are required: %s") %
                       ", ".join(required_actions))

        # make sure all required groups had one option present
        for group in self._mutually_exclusive_groups:
            if group.required:
                for action in group._group_actions:
                    if action in seen_non_default_actions:
                        break

                # if no actions were used, report the error
                else:
                    names: list[str] = []
                    for action in group._group_actions:
                        if (
                                action.help is not SUPPRESS
                        ) and (
                            name := _get_action_name(action)
                        ):
                            names.append(name)
                    msg = _("one of the arguments %s is required")
                    self.error(msg % " ".join(names))

        # return the updated namespace and the extra arguments
        return namespace, extras



import gettext


class PacmanTranslation():

    translation: gettext.NullTranslations | None = None

    @classmethod
    def get(cls) -> gettext.NullTranslations:
        if not cls.translation:
            cls.translation = gettext.translation("pacman", fallback=True)
        return cls.translation


def _p(msg: str) -> str:
    return PacmanTranslation.get().gettext(msg)



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import codecs
import enum
import os
import shutil
import subprocess  # nosec B404
import sys
import tempfile
from multiprocessing.pool import ThreadPool
from time import sleep
from typing import IO, TYPE_CHECKING, Any, Callable, TypeVar

import pyalpm

from .args import parse_args
from .config import RUNNING_AS_ROOT, PikaurConfig
from .i18n import translate
from .pprint import ColorsHighlight, bold_line, color_line, print_error, print_stderr

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from typing_extensions import NotRequired, TypedDict

    from .aur import AURPackageInfo

    class SpawnArgs(TypedDict):
        stdout: NotRequired["IOStream"]
        stderr: NotRequired["IOStream"]
        cwd: NotRequired[str]
        env: NotRequired[dict[str, str]]


DEFAULT_INPUT_ENCODING = "utf-8"
PIPE = subprocess.PIPE
IOStream = IO[bytes] | int | None


class ComparableType:

    __ignore_in_eq__: tuple[str, ...] = ()

    __hash__ = object.__hash__
    __compare_stack__: list["ComparableType"] | None = None

    @property
    def public_values(self) -> dict[str, Any]:
        return {
            var: val for var, val in vars(self).items()
            if not var.startswith("__")
        }

    def __eq__(self, other: "ComparableType") -> bool:  # type: ignore[override]
        if not isinstance(other, self.__class__):
            return False
        if not self.__compare_stack__:
            self.__compare_stack__ = []
        elif other in self.__compare_stack__:
            return super().__eq__(other)
        self.__compare_stack__.append(other)
        self_values = {}
        self_values.update(self.public_values)
        others_values = {}
        others_values.update(other.public_values)
        for values in (self_values, others_values):
            for skip_prop in self.__ignore_in_eq__:
                if skip_prop in values:
                    del values[skip_prop]
        result = self_values == others_values
        self.__compare_stack__ = None
        return result


class DataType(ComparableType):

    ignore_extra_properties: bool

    @property
    def __all_annotations__(self) -> dict[str, type]:
        annotations: dict[str, type] = {}
        for parent_class in reversed(self.__class__.mro()):
            annotations.update(**getattr(parent_class, "__annotations__", {}))
        return annotations

    def _key_exists(self, key: str) -> bool:
        return key in dir(self)

    def __init__(self, *, ignore_extra_properties: bool = False, **kwargs: Any) -> None:
        self.ignore_extra_properties = ignore_extra_properties
        for key, value in kwargs.items():
            setattr(self, key, value)
        for key in self.__all_annotations__:
            if not self._key_exists(key):
                raise TypeError(
                    f"'{self.__class__.__name__}' does "
                    f"not have required attribute '{key}' set"
                )

    def __setattr__(self, key: str, value: Any) -> None:
        if not (
            (
                key in self.__all_annotations__
            ) or self._key_exists(key)
        ):
            if self.ignore_extra_properties:
                print_error(f"Unexpected key `{key}` in {self.__class__.__name__}")
            else:
                raise TypeError(
                    f"'{self.__class__.__name__}' does "
                    f"not have attribute '{key}'"
                )
        super().__setattr__(key, value)


class PackageSource(enum.Enum):
    REPO = enum.auto()
    AUR = enum.auto()
    LOCAL = enum.auto()


class InstallInfo(DataType):
    name: str
    current_version: str
    new_version: str
    description: str | None = None
    maintainer: str | None = None
    repository: str | None = None
    devel_pkg_age_days: int | None = None
    package: "pyalpm.Package | AURPackageInfo"
    provided_by: list["pyalpm.Package | AURPackageInfo"] | None = None
    required_by: list["InstallInfo"] | None = None
    members_of: list[str] | None = None
    replaces: list[str] | None = None
    pkgbuild_path: str | None = None

    __ignore_in_eq__ = ("package", "provided_by", "pkgbuild_path")

    @property
    def package_source(self) -> PackageSource:
        if isinstance(self.package, pyalpm.Package):
            return PackageSource.REPO
        return PackageSource.AUR

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__} "{self.name}" '
            f"{self.current_version} -> {self.new_version}>"
        )


class RepoInstallInfo(InstallInfo):
    package: "pyalpm.Package"


class AURInstallInfo(InstallInfo):
    package: "AURPackageInfo"


def sudo(cmd: list[str]) -> list[str]:
    if running_as_root():
        return cmd
    return [PikaurConfig().misc.PrivilegeEscalationTool.get_str(), ] + cmd


def get_sudo_refresh_command() -> list[str]:
    pacman_path = PikaurConfig().misc.PacmanPath.get_str()
    return sudo([pacman_path, "-T"])


class InteractiveSpawn(subprocess.Popen[bytes]):

    stdout_text: str | None
    stderr_text: str | None

    def communicate(
            self, com_input: bytes | None = None, timeout: float | None = None
    ) -> tuple[bytes, bytes]:
        if parse_args().print_commands:
            if self.args != get_sudo_refresh_command():
                print_stderr(
                    color_line("=> ", ColorsHighlight.cyan) +
                    (
                        " ".join(str(arg) for arg in self.args)
                        if isinstance(self.args, list) else
                        str(self.args)
                    )
                )

        stdout, stderr = super().communicate(com_input, timeout)
        self.stdout_text = stdout.decode(DEFAULT_INPUT_ENCODING) if stdout else None
        self.stderr_text = stderr.decode(DEFAULT_INPUT_ENCODING) if stderr else None
        return stdout, stderr

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} returned {self.returncode}:\n"
            f"STDOUT:\n{self.stdout_text}\n\n"
            f"STDERR:\n{self.stderr_text}"
        )

    def __del__(self) -> None:
        self.terminate()
        self.communicate()


def interactive_spawn(
        cmd: list[str],
        stdout: IOStream = None,
        stderr: IOStream = None,
        cwd: str | None = None,
        env: dict[str, str] | None = None,
) -> InteractiveSpawn:
    kwargs: "SpawnArgs" = {}
    if stdout:
        kwargs["stdout"] = stdout
    if stderr:
        kwargs["stderr"] = stderr
    if cwd:
        kwargs["cwd"] = cwd
    if env:
        kwargs["env"] = env
    process = InteractiveSpawn(
        cmd, **kwargs
    )
    process.communicate()
    return process


def spawn(
        cmd: list[str],
        cwd: str | None = None,
        env: dict[str, str] | None = None,
) -> InteractiveSpawn:
    with tempfile.TemporaryFile() as out_file:
        with tempfile.TemporaryFile() as err_file:
            proc = interactive_spawn(
                cmd, stdout=out_file, stderr=err_file,
                cwd=cwd, env=env,
            )
            out_file.seek(0)
            err_file.seek(0)
            proc.stdout_text = out_file.read().decode(DEFAULT_INPUT_ENCODING)
            proc.stderr_text = err_file.read().decode(DEFAULT_INPUT_ENCODING)
    return proc


def joined_spawn(
        cmd: list[str],
        cwd: str | None = None,
        env: dict[str, str] | None = None,
) -> InteractiveSpawn:
    with tempfile.TemporaryFile() as out_file:
        proc = interactive_spawn(
            cmd, stdout=out_file, stderr=out_file,
            cwd=cwd, env=env,
        )
        out_file.seek(0)
        proc.stdout_text = out_file.read().decode(DEFAULT_INPUT_ENCODING)
    return proc


def running_as_root() -> bool:
    return RUNNING_AS_ROOT


def isolate_root_cmd(
        cmd: list[str],
        cwd: str | None = None,
        env: dict[str, str] | None = None,
) -> list[str]:
    if not running_as_root():
        return cmd
    base_root_isolator = [
        "/usr/sbin/systemd-run",
        "--service-type=oneshot",
        "--pipe", "--wait", "--pty",
        "-p", "DynamicUser=yes",
        "-p", "CacheDirectory=pikaur",
        "-E", "HOME=/tmp",
    ]
    if env is not None:
        for env_var_name, env_var_value in env.items():
            base_root_isolator += ["-E", f"{env_var_name}={env_var_value}"]
    if cwd is not None:
        base_root_isolator += ["-p", "WorkingDirectory=" + os.path.abspath(cwd)]
    for env_var_name in (
            "http_proxy", "https_proxy", "ftp_proxy",
    ):
        if os.environ.get(env_var_name) is not None:
            base_root_isolator += ["-E", f"{env_var_name}={os.environ[env_var_name]}"]
    return base_root_isolator + cmd


def detect_bom_type(file_path: str) -> str:
    """
    returns file encoding string for open() function
    https://stackoverflow.com/a/44295590/1850190
    """
    with open(file_path, "rb") as test_file:
        first_bytes = test_file.read(4)

    if first_bytes[0:3] == b"\xef\xbb\xbf":
        return "utf-8"

    # Python automatically detects endianness if utf-16 bom is present
    # write endianness generally determined by endianness of CPU
    if (
            first_bytes[0:2] == b"\xfe\xff"
    ) or (
        first_bytes[0:2] == b"\xff\xfe"
    ):
        return "utf16"

    if (
            first_bytes[0:5] == b"\xfe\xff\x00\x00"
    ) or (
        first_bytes[0:5] == b"\x00\x00\xff\xfe"
    ):
        return "utf32"

    # If BOM is not provided, then assume its the codepage
    #     used by your operating system
    return "cp1252"
    # For the United States its: cp1252


def open_file(
        file_path: str, mode: str = "r", encoding: str | None = None
) -> codecs.StreamReaderWriter:
    if encoding is None and (mode and "r" in mode):
        encoding = detect_bom_type(file_path)
    if encoding:
        return codecs.open(
            file_path, mode, errors="ignore", encoding=encoding
        )
    return codecs.open(
        file_path, mode, errors="ignore"
    )


def replace_file(src: str, dest: str) -> None:
    if os.path.exists(src):
        if os.path.exists(dest):
            os.remove(dest)
        shutil.move(src, dest)


def remove_dir(dir_path: str) -> None:
    try:
        shutil.rmtree(dir_path)
    except PermissionError:
        interactive_spawn(sudo(["rm", "-rf", dir_path]))


def get_editor() -> list[str] | None:
    editor_line = os.environ.get("VISUAL") or os.environ.get("EDITOR")
    if editor_line:
        return editor_line.split(" ")
    for editor in (
            "vim", "nano", "mcedit", "edit", "emacs", "nvim", "kak",
            "e3", "atom", "adie", "dedit", "gedit", "jedit", "kate", "kwrite", "leafpad",
            "mousepad", "notepadqq", "pluma", "code", "xed", "nvim-qt", "geany",
    ):
        path = shutil.which(editor)
        if path:
            return [path, ]
    return None


def dirname(path: str) -> str:
    return os.path.dirname(path) or "."


def sudo_loop(*, once: bool = False) -> None:
    """Get sudo for further questions."""
    sudo_loop_interval = PikaurConfig().misc.SudoLoopInterval.get_int()
    if sudo_loop_interval == -1:
        return
    while True:
        interactive_spawn(get_sudo_refresh_command())
        if once:
            break
        sleep(sudo_loop_interval)


SudoLoopResultT = TypeVar("SudoLoopResultT")


def run_with_sudo_loop(function: Callable[..., SudoLoopResultT]) -> SudoLoopResultT | None:
    sudo_loop(once=True)
    with ThreadPool(processes=2) as pool:
        main_thread = pool.apply_async(function, ())
        pool.apply_async(sudo_loop)
        pool.close()
        catched_exc = None
        result: SudoLoopResultT | None = None
        try:
            result = main_thread.get()
        except Exception as exc:
            catched_exc = exc
        finally:
            pool.terminate()
        if catched_exc:
            raise catched_exc
        return result


def check_systemd_dynamic_users() -> bool:  # pragma: no cover
    try:
        out = subprocess.check_output(["/usr/sbin/systemd-run", "--version"],  # nosec B603
                                      universal_newlines=True)
    except FileNotFoundError:
        return False
    first_line = out.split("\n", maxsplit=1)[0]
    version = int(first_line.split(maxsplit=2)[1])
    return version >= 235


def check_runtime_deps(dep_names: list[str] | None = None) -> None:
    if sys.version_info < (3, 7):
        print_error(
            translate("pikaur requires Python >= 3.7 to run."),
        )
        sys.exit(65)
    if running_as_root() and not check_systemd_dynamic_users():
        print_error(
            translate("pikaur requires systemd >= 235 (dynamic users) to be run as root."),
        )
        sys.exit(65)
    if not dep_names:
        privilege_escalation_tool = PikaurConfig().misc.PrivilegeEscalationTool.get_str()
        dep_names = ["fakeroot", ] + (
            [privilege_escalation_tool] if not running_as_root() else []
        )

    for dep_bin in dep_names:
        if not shutil.which(dep_bin):
            print_error("'{}' {}.".format(  # pylint: disable=consider-using-f-string
                bold_line(dep_bin),
                translate("executable not found")
            ))
            sys.exit(2)



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import configparser
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

from .i18n import translate

if TYPE_CHECKING:
    from typing_extensions import NotRequired, TypedDict

    class DeprecatedConfigValue(TypedDict):
        section: str
        option: str
        transform: NotRequired[Callable[[str, configparser.ConfigParser], str]]

    class ConfigValueType(TypedDict):
        data_type: str
        default: NotRequired[str]
        deprecated: NotRequired[DeprecatedConfigValue]
        migrated: NotRequired[bool]


DEFAULT_CONFIG_ENCODING = "utf-8"


RUNNING_AS_ROOT = os.geteuid() == 0  # @TODO: could global var be avoided here?


VERSION = '1.14.6-dev'

_USER_CACHE_HOME = os.environ.get(
    "XDG_CACHE_HOME",
    os.path.join(Path.home(), ".cache/")
)
if RUNNING_AS_ROOT:
    CACHE_ROOT = "/var/cache/pikaur"
else:
    CACHE_ROOT = os.path.join(_USER_CACHE_HOME, "pikaur/")

BUILD_CACHE_PATH = os.path.join(CACHE_ROOT, "build")
PACKAGE_CACHE_PATH = os.path.join(CACHE_ROOT, "pkg")

CONFIG_ROOT = os.environ.get(
    "XDG_CONFIG_HOME",
    os.path.join(Path.home(), ".config/")
)

DATA_ROOT = os.path.join(
    os.environ.get(
        "XDG_DATA_HOME",
        os.path.join(Path.home(), ".local/share/")
    ), "pikaur"
)
_OLD_AUR_REPOS_CACHE_PATH = os.path.join(CACHE_ROOT, "aur_repos")
if RUNNING_AS_ROOT:
    AUR_REPOS_CACHE_PATH = os.path.join(CACHE_ROOT, "aur_repos")
else:
    AUR_REPOS_CACHE_PATH = os.path.join(DATA_ROOT, "aur_repos")

if RUNNING_AS_ROOT:
    BUILD_DEPS_LOCK = os.path.join(_USER_CACHE_HOME, "pikaur_build_deps.lock")
else:
    BUILD_DEPS_LOCK = "/tmp/pikaur_build_deps.lock"  # nosec B108


def get_config_path() -> str:
    config_flag = "--pikaur-config"
    if config_flag in sys.argv:
        return sys.argv[
            sys.argv.index(config_flag) + 1
        ]
    return os.path.join(
        CONFIG_ROOT,
        "pikaur.conf"
    )


CONFIG_SCHEMA: dict[str, dict[str, "ConfigValueType"]] = {
    "sync": {
        "AlwaysShowPkgOrigin": {
            "data_type": "bool",
            "default": "no",
        },
        "DevelPkgsExpiration": {
            "data_type": "int",
            "default": "-1",
        },
        "UpgradeSorting": {
            "data_type": "str",
            "default": "versiondiff"
        },
        "ShowDownloadSize": {
            "data_type": "bool",
            "default": "no",
        },
        "IgnoreOutofdateAURUpgrades": {
            "data_type": "bool",
            "default": "no",
        },
    },
    "build": {
        "KeepBuildDir": {
            "data_type": "bool",
            "default": "no",
        },
        "KeepDevBuildDir": {
            "data_type": "bool",
            "default": "yes",
        },
        "KeepBuildDeps": {
            "data_type": "bool",
            "default": "no",
        },
        "GpgDir": {
            "data_type": "str",
            "default": ("/etc/pacman.d/gnupg/" if RUNNING_AS_ROOT else "")
        },
        "SkipFailedBuild": {
            "data_type": "bool",
            "default": "no",
        },
        "AlwaysUseDynamicUsers": {
            "data_type": "bool",
            "default": "no",
        },
        "NoEdit": {
            "data_type": "bool",
            "deprecated": {
                "section": "review",
                "option": "NoEdit",
            },
        },
        "DontEditByDefault": {
            "data_type": "bool",
            "deprecated": {
                "section": "review",
                "option": "DontEditByDefault",
            },
        },
        "NoDiff": {
            "data_type": "bool",
            "deprecated": {
                "section": "review",
                "option": "NoDiff",
            },
        },
        "GitDiffArgs": {
            "data_type": "str",
            "deprecated": {
                "section": "review",
                "option": "GitDiffArgs",
            },
        },
        "IgnoreArch": {
            "data_type": "bool",
            "default": "no"
        },
    },
    "review": {
        "NoEdit": {
            "data_type": "bool",
            "default": "no",
        },
        "DontEditByDefault": {
            "data_type": "bool",
            "default": "no",
        },
        "NoDiff": {
            "data_type": "bool",
            "default": "no",
        },
        "GitDiffArgs": {
            "data_type": "str",
            "default": "--ignore-space-change,--ignore-all-space",
        },
        "DiffPager": {
            "data_type": "str",
            "default": "auto"
        },
        "HideDiffFiles": {
            "data_type": "str",
            "default": ".SRCINFO"
        },
    },
    "colors": {
        "Version": {
            "data_type": "int",
            "default": "10",
        },
        "VersionDiffOld": {
            "data_type": "int",
            "default": "11",
        },
        "VersionDiffNew": {
            "data_type": "int",
            "default": "9",
        },
    },
    "ui": {
        "RequireEnterConfirm": {
            "data_type": "bool",
            "default": "yes"
        },
        "DiffPager": {
            "data_type": "str",
            "deprecated": {
                "section": "review",
                "option": "DiffPager",
            },
        },
        "PrintCommands": {
            "data_type": "bool",
            "default": "no"
        },
        "AurSearchSorting": {
            "data_type": "str",
            "default": "hottest"
        },
        "DisplayLastUpdated": {
            "data_type": "bool",
            "default": "no"
        },
        "GroupByRepository": {
            "data_type": "bool",
            "default": "yes"
        },
        "ReverseSearchSorting": {
            "data_type": "bool",
            "default": "no"
        },
        "WarnAboutPackageUpdates": {
            "data_type": "str",
            "default": ""
        },
    },
    "misc": {
        "SudoLoopInterval": {
            "data_type": "int",
            "default": "59",
        },
        "PacmanPath": {
            "data_type": "str",
            "default": "pacman"
        },
        "PrivilegeEscalationTool": {
            "data_type": "str",
            "default": "sudo",
        },
        "AurHost": {
            "data_type": "str",
            "deprecated": {
                "section": "network",
                "option": "AurUrl",
                "transform": lambda old_value, _config: f"https://{old_value}"
            },
        },
        "NewsUrl": {
            "data_type": "str",
            "deprecated": {
                "section": "network",
                "option": "NewsUrl",
            },
        },
    },
    "network": {
        "AurUrl": {
            "data_type": "str",
            "default": "https://aur.archlinux.org",
        },
        "NewsUrl": {
            "data_type": "str",
            "default": "https://www.archlinux.org/feeds/news/",
        },
        "Socks5Proxy": {
            "data_type": "str",
            "default": "",
        },
        "AurHttpProxy": {
            "data_type": "str",
            "default": "",
        },
        "AurHttpsProxy": {
            "data_type": "str",
            "default": "",
        },
    },
}


def get_key_type(section_name: str, key_name: str) -> str | None:
    config_value: "ConfigValueType" | None = CONFIG_SCHEMA.get(section_name, {}).get(key_name, None)
    if not config_value:
        return None
    return config_value.get("data_type")


def write_config(config: configparser.ConfigParser | None = None) -> None:
    if not config:
        config = configparser.ConfigParser()
    need_write = False
    for section_name, section in CONFIG_SCHEMA.items():
        if section_name not in config:
            config[section_name] = {}
        for option_name, option_schema in section.items():
            if option_schema.get("migrated"):
                need_write = True
                continue
            if option_schema.get("deprecated"):
                continue
            if option_name not in config[section_name]:
                config[section_name][option_name] = option_schema["default"]
                need_write = True
    if need_write:
        if not os.path.exists(CONFIG_ROOT):
            os.makedirs(CONFIG_ROOT)
        with open(get_config_path(), "w", encoding=DEFAULT_CONFIG_ENCODING) as configfile:
            config.write(configfile)


def str_to_bool(value: str) -> bool:
    # pylint:disable=protected-access
    result: bool = \
        configparser.RawConfigParser()._convert_to_boolean(value)  # type: ignore[attr-defined]
    return result


class PikaurConfigItem:

    def __init__(self, section: configparser.SectionProxy, key: str) -> None:
        self.section = section
        self.key = key
        self.value = self.section.get(key)

    def get_bool(self) -> bool:
        if get_key_type(self.section.name, self.key) != "bool":
            raise TypeError(f"{self.key} is not 'bool'")
        return str_to_bool(self.value)

    def get_int(self) -> int:
        if get_key_type(self.section.name, self.key) != "int":
            raise TypeError(f"{self.key} is not 'int'")
        return int(self.value)

    def get_str(self) -> str:
        # note: it's basically needed for mypy
        if get_key_type(self.section.name, self.key) != "str":
            raise TypeError(f"{self.key} is not 'str'")
        return str(self.value)

    def __str__(self) -> str:
        return self.get_str()

    def __eq__(self, item: Any) -> bool:
        result: bool = self.get_str() == item
        return result


class PikaurConfigSection():

    section: configparser.SectionProxy

    def __init__(self, section: configparser.SectionProxy) -> None:
        self.section = section

    def __getattr__(self, attr: str) -> PikaurConfigItem:
        return PikaurConfigItem(self.section, attr)

    def __repr__(self) -> str:
        return str(self.section)


class PikaurConfig():

    _config: configparser.ConfigParser

    @classmethod
    def get_config(cls) -> configparser.ConfigParser:
        if not getattr(cls, "_config", None):
            config_path = get_config_path()
            cls._config = configparser.ConfigParser()
            if not os.path.exists(config_path):
                write_config()
            cls._config.read(config_path, encoding=DEFAULT_CONFIG_ENCODING)
            cls.migrate_config()
            write_config(config=cls._config)
            cls.validate_config()
        return cls._config

    @classmethod
    def migrate_config(cls) -> None:
        for section_name, section in CONFIG_SCHEMA.items():
            for option_name, option_schema in section.items():
                if not option_schema.get("deprecated"):
                    continue

                new_section_name: str = option_schema["deprecated"]["section"]
                new_option_name: str = option_schema["deprecated"]["option"]
                transform: Callable[
                    [str, configparser.ConfigParser], str
                ] | None = option_schema["deprecated"].get("transform")

                old_value_was_migrated = False
                if (
                        new_section_name not in cls._config
                ) or (
                    cls._config[new_section_name].get(new_option_name) is None
                ):
                    value_to_migrate = cls._config[section_name].get(option_name)
                    if value_to_migrate is not None:
                        if transform:
                            value_to_migrate = transform(value_to_migrate, cls._config)
                        if new_section_name not in cls._config:
                            cls._config[new_section_name] = {}
                        cls._config[new_section_name][new_option_name] = value_to_migrate
                        CONFIG_SCHEMA[new_section_name][new_option_name]["migrated"] = True
                        old_value_was_migrated = True

                old_value_was_removed = False
                if option_name in cls._config[section_name]:
                    del cls._config[section_name][option_name]
                    CONFIG_SCHEMA[section_name][option_name]["migrated"] = True
                    old_value_was_removed = True

                if old_value_was_migrated or old_value_was_removed:
                    print(" ".join([  # noqa: T201
                        "::",
                        translate("warning:"),
                        translate('Migrating [{}]{} config option to [{}]{} = "{}"...').format(
                            section_name, option_name,
                            new_section_name, new_option_name,
                            cls._config[new_section_name][new_option_name]
                        ),
                        "\n",
                    ]))

    @classmethod
    def validate_config(cls) -> None:
        pacman_path = cls._config["misc"]["PacmanPath"]
        if pacman_path == "pikaur" or pacman_path.endswith("/pikaur"):
            print("BAM! I am a shell bomb.")  # noqa: T201
            sys.exit(1)

    def __getattr__(self, attr: str) -> PikaurConfigSection:
        return PikaurConfigSection(self.get_config()[attr])



from threading import Lock
from typing import Any


class FancyLock():

    _fancy_lock: "Lock | None" = None

    @classmethod
    def get_lock(cls) -> Lock:
        if not cls._fancy_lock:
            cls._fancy_lock = Lock()
        return cls._fancy_lock

    @property
    def fancy_lock(self) -> Lock:
        return self.get_lock()

    def __enter__(self) -> None:
        self.fancy_lock.acquire()

    def __exit__(self, *_exc_details: Any) -> None:
        if self.fancy_lock.locked():
            self.fancy_lock.release()

    def __del__(self) -> None:
        self.__exit__()



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import os
from typing import TypeVar

from .args import parse_args
from .config import CONFIG_ROOT
from .core import open_file, running_as_root

ConfigValueType = str | list[str] | None
ConfigFormat = dict[str, ConfigValueType]


FallbackValueT = TypeVar("FallbackValueT")


class ConfigReader():

    comment_prefixes = ("#", ";")

    _cached_config: dict[str, ConfigFormat] | None = None
    default_config_path: str
    list_fields: list[str] = []
    ignored_fields: list[str] = []

    @classmethod
    def _parse_line(cls, line: str) -> tuple[str | None, ConfigValueType]:
        blank = (None, None, )
        if line.startswith(" "):
            return blank
        if "=" not in line:
            return blank
        line = line.strip()
        for comment_prefix in cls.comment_prefixes:
            line, *_comments = line.split(comment_prefix)

        key, _sep, value = line.partition("=")
        key = key.strip()
        value = value.strip()

        if key in cls.ignored_fields:
            return blank

        if value:
            value = value.strip('"').strip("'")
        else:
            return key, value

        if key in cls.list_fields:
            list_value = value.split()
            return key, list_value

        return key, value

    @classmethod
    def get_config(cls, config_path: str | None = None) -> ConfigFormat:
        config_path = config_path or cls.default_config_path
        if cls._cached_config is None:
            cls._cached_config = {}
        if not cls._cached_config.get(config_path):
            with open_file(config_path) as config_file:
                # pylint: disable=unsupported-assignment-operation
                cls._cached_config[config_path] = {
                    key: value
                    for key, value in [
                        cls._parse_line(line)
                        for line in config_file.readlines()
                    ] if key
                }
        # pylint: disable=unsubscriptable-object
        return cls._cached_config[config_path]

    @classmethod
    def get(
            cls,
            key: str,
            fallback: FallbackValueT | None = None,
            config_path: str | None = None
    ) -> ConfigValueType | FallbackValueT:
        return cls.get_config(config_path=config_path).get(key) or fallback


class MakepkgConfig():

    _user_makepkg_path: str | None = "unset"

    @classmethod
    def get_user_makepkg_path(cls) -> str | None:
        if cls._user_makepkg_path == "unset":
            possible_paths = [
                os.path.expanduser("~/.makepkg.conf"),
                os.path.join(CONFIG_ROOT, "pacman/makepkg.conf"),
            ]
            config_path: str | None = None
            for path in possible_paths:
                if os.path.exists(path):
                    config_path = path
            cls._user_makepkg_path = config_path
        return cls._user_makepkg_path

    @classmethod
    def get(
            cls,
            key: str,
            fallback: FallbackValueT | None = None,
            config_path: str | None = None
    ) -> ConfigValueType | FallbackValueT:
        arg_path: str | None = parse_args().makepkg_config
        value: ConfigValueType | FallbackValueT = ConfigReader.get(
            key, fallback, config_path="/etc/makepkg.conf"
        )
        if cls.get_user_makepkg_path():
            value = ConfigReader.get(key, value, config_path=cls.get_user_makepkg_path())
        if arg_path:
            value = ConfigReader.get(key, value, config_path=arg_path)
        if config_path:
            value = ConfigReader.get(key, value, config_path=config_path)
        return value


def get_pkgdest() -> str | None:
    config_pkgdest = MakepkgConfig.get("PKGDEST")
    if not isinstance(config_pkgdest, str):
        config_pkgdest = None
    pkgdest: str | None = os.environ.get("PKGDEST", config_pkgdest)
    if pkgdest:
        pkgdest = pkgdest.replace("$HOME", "~")
        pkgdest = os.path.expanduser(pkgdest)
    return pkgdest


class MakePkgCommand:

    _cmd: list[str] | None = None
    pkgdest_skipped = False

    @classmethod
    def _apply_dynamic_users_workaround(cls) -> None:
        pkgdest = get_pkgdest()
        if running_as_root() and pkgdest and (
                pkgdest.startswith("/tmp") or  # nosec B108
                pkgdest.startswith("/var/tmp")  # nosec B108
        ):
            if not cls._cmd:
                raise RuntimeError()
            cls._cmd = ["env", "PKGDEST="] + cls._cmd
            cls.pkgdest_skipped = True

    @classmethod
    def get(cls) -> list[str]:
        if cls._cmd is None:
            args = parse_args()
            makepkg_flags = (
                args.mflags.split(",") if args.mflags else []
            )
            config_args = (
                ["--config", args.makepkg_config] if args.makepkg_config else []
            )
            cls._cmd = [args.makepkg_path or "makepkg", ] + makepkg_flags + config_args
            cls._apply_dynamic_users_workaround()
        return cls._cmd



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

from typing import TYPE_CHECKING

from .core import DataType, PackageSource

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .build import PackageBuild
    from .core import InteractiveSpawn
    from .version import VersionMatcher


class PackagesNotFoundError(DataType, Exception):
    packages: list[str]
    wanted_by: list[str] | None = None

    def __init__(self, packages: list[str], wanted_by: list[str] | None = None) -> None:
        DataType.__init__(self, packages=packages, wanted_by=wanted_by)
        message = ", ".join(packages)
        if wanted_by:
            message += f" wanted by {', '.join(wanted_by)}"
        Exception.__init__(self, message)


class PackagesNotFoundInRepoError(PackagesNotFoundError):
    # pass
    # @TODO: pylint bug:
    packages: list[str]


class PackagesNotFoundInAURError(PackagesNotFoundError):
    # pass
    # @TODO: pylint bug:
    packages: list[str]


class BuildError(Exception):
    pass


class CloneError(DataType, Exception):
    build: "PackageBuild"
    result: "InteractiveSpawn"


class DependencyError(Exception):
    pass


class DependencyVersionMismatchError(DataType, Exception):
    version_found: dict[str, str] | str
    dependency_line: str
    who_depends: str
    depends_on: str
    location: PackageSource
    version_matcher: "VersionMatcher | None" = None

    def __init__(  # pylint: disable=too-many-arguments
            self,
            version_found: dict[str, str] | str,
            dependency_line: str,
            who_depends: str,
            depends_on: str,
            location: PackageSource,
            version_matcher: "VersionMatcher | None" = None,
    ) -> None:
        super().__init__(
            version_found=version_found,
            dependency_line=dependency_line,
            who_depends=who_depends,
            depends_on=depends_on,
            location=location,
            version_matcher=version_matcher,
        )
        if self.version_matcher:
            self.dependency_line = self.version_matcher.line


class DependencyNotBuiltYetError(Exception):
    pass


class AURError(Exception):
    url: str
    error: str

    def __init__(self, url: str, error: str) -> None:
        self.url = url
        self.error = error
        super().__init__(f"URL: {self.url}\nError: {self.error}")


class SysExit(Exception):  # noqa: N818
    code: int

    def __init__(self, code: int) -> None:
        self.code = code
        super().__init__(f"Exit code: {code}")



import os
from typing import TYPE_CHECKING, Callable

from .args import parse_args, reconstruct_args
from .config import BUILD_CACHE_PATH, PACKAGE_CACHE_PATH, PikaurConfig
from .core import interactive_spawn, remove_dir, sudo
from .exceptions import SysExit
from .i18n import translate
from .pikspect import YesNo, format_pacman_question, pikspect
from .pprint import ColorsHighlight, bold_line, color_line, create_debug_logger, print_stdout
from .prompt import ask_to_continue

if TYPE_CHECKING:
    from subprocess import Popen  # nosec B404


_debug = create_debug_logger("pkg_cache_cli")


def clean_aur_cache() -> None:
    args = parse_args()
    for directory, message, minimal_clean_level in (
            (BUILD_CACHE_PATH, translate("Build directory"), 1, ),
            (PACKAGE_CACHE_PATH, translate("Packages directory"), 2, ),
    ):
        print_stdout(f"\n{message}: {directory}")
        if minimal_clean_level <= args.clean and os.path.exists(directory):
            if ask_to_continue(text="{} {}".format(  # pylint: disable=consider-using-f-string
                    color_line("::", ColorsHighlight.blue),
                    bold_line(translate("Do you want to remove all files?"))
            )):
                print_stdout(translate("removing all files from cache..."))
                remove_dir(directory)
        else:
            print_stdout(translate("Directory is empty."))


def clean_repo_cache() -> None:
    args = parse_args()
    spawn_func: Callable[[list[str]], "Popen[bytes]"] = interactive_spawn
    if args.noconfirm:

        def noconfirm_cache_remove(pacman_args: list[str]) -> "Popen[bytes]":
            return pikspect(pacman_args, extra_questions={YesNo.ANSWER_Y: [
                format_pacman_question(
                    "Do you want to remove ALL files from cache?",
                    question=YesNo.QUESTION_YN_NO,
                ),
                format_pacman_question(
                    "Do you want to remove all other packages from cache?"
                ),
                format_pacman_question(
                    "Do you want to remove unused repositories?",
                ),
            ]})
        spawn_func = noconfirm_cache_remove
    raise SysExit(
        spawn_func(sudo(
            [PikaurConfig().misc.PacmanPath.get_str(), ] +
            reconstruct_args(args, ignore_args=["noconfirm", ])
        )).returncode
    )


def cli_clean_packages_cache() -> None:
    args = parse_args()
    if not args.repo:
        clean_aur_cache()
    if not args.aur:
        clean_repo_cache()



from datetime import datetime
from multiprocessing.pool import ThreadPool
from unicodedata import east_asian_width

from .args import parse_args, reconstruct_args
from .aur import find_aur_packages, get_all_aur_names
from .core import spawn
from .i18n import translate
from .pacman import get_pacman_command, refresh_pkg_db_if_needed
from .pacman_i18n import _p
from .pprint import ColorsHighlight, bold_line, color_line, print_stdout


def _info_packages_thread_repo() -> str | None:
    args = parse_args()
    proc = spawn(
        get_pacman_command() + reconstruct_args(args, ignore_args=["refresh"]) + args.positional
    )
    return proc.stdout_text


def get_info_fields() -> dict[str, str]:
    return {
        "git_url": translate("AUR Git URL"),
        # "aur_id": translate("id"),
        "name": translate("Name"),
        # packagebaseid=translate(""),
        "packagebase": translate("Package Base"),
        "version": translate("Version"),
        "desc": translate("Description"),
        "url": translate("URL"),
        "keywords": translate("Keywords"),
        "pkg_license": translate("Licenses"),
        "groups": translate("Groups"),
        "provides": translate("Provides"),
        "depends": translate("Depends On"),
        "optdepends": translate("Optional Deps"),
        "makedepends": translate("Make Deps"),
        "checkdepends": translate("Check Deps"),
        "conflicts": translate("Conflicts With"),
        "replaces": translate("Replaces"),
        "submitter": translate("Submitter"),
        "maintainer": translate("Maintainer"),
        "comaintainers": translate("Co-maintainers"),
        "numvotes": translate("Votes"),
        "popularity": translate("Popularity"),
        "firstsubmitted": translate("First Submitted"),
        "lastmodified": translate("Last Updated"),
        "outofdate": translate("Out-of-date"),
    }


def _decorate_repo_info_output(output: str) -> str:
    return output.replace(
        _p("None"), color_line(_p("None"), ColorsHighlight.black)
    )


def _decorate_aur_info_output(output: str) -> str:
    return output.replace(
        translate("None"), color_line(translate("None"), ColorsHighlight.black)
    )


def cli_info_packages() -> None:  # pylint: disable=too-many-locals
    refresh_pkg_db_if_needed()

    args = parse_args()
    aur_pkg_names = args.positional or get_all_aur_names()
    with ThreadPool() as pool:
        aur_thread = pool.apply_async(find_aur_packages, (aur_pkg_names, ))
        repo_thread = pool.apply_async(_info_packages_thread_repo, ())
        pool.close()
        pool.join()
        repo_result = repo_thread.get()
        aur_result = aur_thread.get()

    if repo_result:
        print_stdout(_decorate_repo_info_output(repo_result), end="")

    aur_pkgs = aur_result[0]
    num_found = len(aur_pkgs)
    info_fields = get_info_fields()
    longest_field_length = max(len(field) for field in info_fields.values())
    for i, aur_pkg in enumerate(aur_pkgs):
        pkg_info_lines = []
        for key, display_name in info_fields.items():
            value = getattr(aur_pkg, key, None)
            if key in ["firstsubmitted", "lastmodified", "outofdate"] and value:
                value = datetime.fromtimestamp(value).strftime("%c")
            elif isinstance(value, list):
                value = ", ".join(value) or translate("None")
            key_display = bold_line(_rightpad(display_name, longest_field_length + 1))
            pkg_info_lines.append(f"{key_display}: {value}")
        print_stdout(
            _decorate_aur_info_output("\n".join(pkg_info_lines)) +
            ("\n" if i + 1 < num_found else "")
        )


def _rightpad(text: str, num: int) -> str:
    space = num
    for i in text:
        if east_asian_width(i) in ["F", "W", ]:
            space -= 2
        else:
            space -= 1
    return text + " " * space



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import fnmatch
import re
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Pattern

import pyalpm
from pycman.config import PacmanConfig as PycmanConfig

from .args import PACMAN_APPEND_OPTS, PACMAN_STR_OPTS, parse_args, reconstruct_args
from .config import PikaurConfig
from .core import DataType, PackageSource, spawn, sudo
from .exceptions import DependencyError, PackagesNotFoundInRepoError
from .i18n import translate
from .lock import FancyLock
from .pacman_i18n import _p
from .pprint import color_enabled, create_debug_logger, print_error, print_stderr
from .prompt import retry_interactive_command, retry_interactive_command_or_exit
from .version import VersionMatcher

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .aur import AURPackageInfo


OFFICIAL_REPOS = (
    "core",
    "extra",
    "community",
    "multilib",
    "testing",
    "community-testing",
    "multilib-testing",
)


_debug = create_debug_logger("pacman")


def create_pacman_pattern(pacman_message: str) -> Pattern[str]:
    return re.compile(
        _p(pacman_message).replace(
            "(", r"\("
        ).replace(
            ")", r"\)"
        ).replace(
            "%d", "(.*)"
        ).replace(
            "%s", "(.*)"
        ).replace(
            "%zu", "(.*)"
        ).strip()
    )


def get_pacman_command(ignore_args: list[str] | None = None) -> list[str]:
    ignore_args = ignore_args or []
    args = parse_args()
    pacman_path = PikaurConfig().misc.PacmanPath.get_str()
    pacman_cmd = [pacman_path, ]
    if color_enabled():
        pacman_cmd += ["--color=always"]
    else:
        pacman_cmd += ["--color=never"]

    for _short, arg, _default in PACMAN_STR_OPTS:
        if arg in ["color", ]:  # we force it anyway
            continue
        if arg in ignore_args:
            continue
        value = getattr(args, arg)
        if value:
            pacman_cmd += ["--" + arg, value]

    for _short, arg, _default in PACMAN_APPEND_OPTS:
        if arg in ["ignore", ]:  # we reprocess it anyway
            continue
        if arg in ignore_args:
            continue
        for value in getattr(args, arg) or []:
            pacman_cmd += ["--" + arg, value]

    return pacman_cmd


class PacmanPrint(DataType):

    full_name: str
    repo: str
    name: str


class PacmanConfig(PycmanConfig):

    def __init__(self) -> None:
        super().__init__(conf=parse_args().config or "/etc/pacman.conf")


class ProvidedDependency(DataType):
    name: str
    package: pyalpm.Package
    version_matcher: VersionMatcher

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__} "{self.name}" '
            f"{self.version_matcher.line}>"
        )


def get_pkg_id(pkg: "AURPackageInfo | pyalpm.Package") -> str:
    if isinstance(pkg, pyalpm.Package):
        return f"{pkg.db.name}/{pkg.name}"
    return f"aur/{pkg.name}"


class DbLockRepo(FancyLock):
    pass


class DbLockLocal(FancyLock):
    pass


def get_db_lock(package_source: PackageSource) -> type[DbLockRepo] | type[DbLockLocal]:
    return DbLockRepo if package_source is PackageSource.REPO else DbLockLocal


class PackageDBCommon(metaclass=ABCMeta):

    _packages_list_cache: dict[PackageSource, list[pyalpm.Package]] = {}
    _packages_dict_cache: dict[PackageSource, dict[str, pyalpm.Package]] = {}
    _provided_list_cache: dict[PackageSource, list[str]] = {}
    _provided_dict_cache: dict[PackageSource, dict[str, list[ProvidedDependency]]] = {}

    @classmethod
    def _discard_cache(
            cls, package_source: PackageSource
    ) -> None:
        with get_db_lock(package_source)():
            if cls._packages_list_cache.get(package_source):
                del cls._packages_list_cache[package_source]
            if cls._packages_dict_cache.get(package_source):
                del cls._packages_dict_cache[package_source]
            if cls._provided_list_cache.get(package_source):
                del cls._provided_list_cache[package_source]
            if cls._provided_dict_cache.get(package_source):
                del cls._provided_dict_cache[package_source]

    @classmethod
    def discard_local_cache(cls) -> None:
        cls._discard_cache(PackageSource.LOCAL)

    @classmethod
    def discard_repo_cache(cls) -> None:
        cls._discard_cache(PackageSource.REPO)

    @classmethod
    @abstractmethod
    def get_repo_list(cls, *, quiet: bool = False) -> list[pyalpm.Package]:
        pass
        # if not cls._packages_list_cache.get(PackageSource.REPO):
            # cls._packages_list_cache[PackageSource.REPO] = list(
                # cls.get_repo_dict(quiet=quiet).values()
            # )
        # return cls._packages_list_cache[PackageSource.REPO]

    @classmethod
    @abstractmethod
    def get_local_list(cls, *, quiet: bool = False) -> list[pyalpm.Package]:
        pass
        # if not cls._packages_list_cache.get(PackageSource.LOCAL):
            # cls._packages_list_cache[PackageSource.LOCAL] = list(
                # cls.get_local_dict(quiet=quiet).values()
            # )
        # return cls._packages_list_cache[PackageSource.LOCAL]

    @classmethod
    def get_repo_dict(cls, *, quiet: bool = False) -> dict[str, pyalpm.Package]:
        if not cls._packages_dict_cache.get(PackageSource.REPO):
            cls._packages_dict_cache[PackageSource.REPO] = {
                get_pkg_id(pkg): pkg
                for pkg in cls.get_repo_list(quiet=quiet)
            }
        return cls._packages_dict_cache[PackageSource.REPO]

    @classmethod
    def get_local_dict(cls, *, quiet: bool = False) -> dict[str, pyalpm.Package]:
        if not cls._packages_dict_cache.get(PackageSource.LOCAL):
            cls._packages_dict_cache[PackageSource.LOCAL] = {
                pkg.name: pkg
                for pkg in cls.get_local_list(quiet=quiet)
            }
        return cls._packages_dict_cache[PackageSource.LOCAL]

    @classmethod
    def _get_provided_dict(
            cls, package_source: PackageSource
    ) -> dict[str, list[ProvidedDependency]]:

        if not cls._provided_dict_cache.get(package_source):
            provided_pkg_names: dict[str, list[ProvidedDependency]] = {}
            for pkg in (
                    cls.get_local_list() if package_source == PackageSource.LOCAL
                    else cls.get_repo_list()
            ):
                provided_pkg_names.setdefault(pkg.name, []).append(
                    ProvidedDependency(
                        name=pkg.name,
                        package=pkg,
                        version_matcher=VersionMatcher(pkg.name, is_pkg_deps=True)
                    )
                )
                if pkg.provides:
                    for provided_pkg_line in pkg.provides:
                        version_matcher = VersionMatcher(provided_pkg_line, is_pkg_deps=True)
                        provided_name = version_matcher.pkg_name
                        provided_pkg_names.setdefault(provided_name, []).append(
                            ProvidedDependency(
                                name=provided_name,
                                package=pkg,
                                version_matcher=version_matcher
                            )
                        )
            for what_provides, provided_pkgs in list(provided_pkg_names.items()):
                if len(provided_pkgs) == 1 and provided_pkgs[0].name == what_provides:
                    del provided_pkg_names[what_provides]
            cls._provided_dict_cache[package_source] = provided_pkg_names
        return cls._provided_dict_cache[package_source]

    @classmethod
    def get_repo_provided_dict(cls) -> dict[str, list[ProvidedDependency]]:
        return cls._get_provided_dict(PackageSource.REPO)

    @classmethod
    def get_local_provided_dict(cls) -> dict[str, list[ProvidedDependency]]:
        return cls._get_provided_dict(PackageSource.LOCAL)

    @classmethod
    def get_repo_pkgnames(cls) -> list[str]:
        return [pkg.name for pkg in cls.get_repo_list()]

    @classmethod
    def get_local_pkgnames(cls) -> list[str]:
        return [pkg.name for pkg in cls.get_local_list()]


class RepositoryNotFoundError(Exception):
    pass


class PackageDB(PackageDBCommon):

    _alpm_handle: pyalpm.Handle | None = None

    _pacman_find_cache: dict[str, list[PacmanPrint]] = {}
    _pacman_test_cache: dict[str, list[VersionMatcher]] = {}
    _pacman_repo_pkg_present_cache: dict[str, bool] = {}

    @classmethod
    def get_alpm_handle(cls) -> pyalpm.Handle:
        if not cls._alpm_handle:
            cls._alpm_handle = PacmanConfig().initialize_alpm()
        if not cls._alpm_handle:
            raise Exception("Cannot initialize ALPM")
        return cls._alpm_handle

    @classmethod
    def discard_local_cache(cls) -> None:
        super().discard_local_cache()
        cls._alpm_handle = None
        cls._pacman_test_cache = {}

    @classmethod
    def discard_repo_cache(cls) -> None:
        super().discard_repo_cache()
        cls._alpm_handle = None
        cls._pacman_find_cache = {}
        cls._pacman_repo_pkg_present_cache = {}

    @classmethod
    def search_local(cls, search_query: str) -> list[pyalpm.Package]:
        return cls.get_alpm_handle().get_localdb().search(search_query)

    @classmethod
    def get_local_list(cls, *, quiet: bool = False) -> list[pyalpm.Package]:
        if not cls._packages_list_cache.get(PackageSource.LOCAL):
            with DbLockLocal():
                if not quiet:
                    print_stderr(translate("Reading local package database..."))
                cls._packages_list_cache[PackageSource.LOCAL] = cls.search_local("")
        return cls._packages_list_cache[PackageSource.LOCAL]

    @classmethod
    def get_repo_priority(cls, repo_name: str) -> int:
        """0 is the highest priority."""
        repos = [r.name for r in cls.get_alpm_handle().get_syncdbs()]
        if repo_name not in repos:
            raise RepositoryNotFoundError(f"'{repo_name}' in {repos}")
        return repos.index(repo_name)

    @classmethod
    def _get_provided_dict(
            cls, package_source: PackageSource
    ) -> dict[str, list[ProvidedDependency]]:

        if not cls._provided_dict_cache.get(package_source):
            provided_pkg_names = super()._get_provided_dict(package_source)
            if package_source == PackageSource.REPO:
                for _what_provides, provided_pkgs in provided_pkg_names.items():
                    provided_pkgs.sort(
                        key=lambda p: f"{cls.get_repo_priority(p.package.db.name)}{p.package.name}"
                    )
            cls._provided_dict_cache[package_source] = provided_pkg_names
        return cls._provided_dict_cache[package_source]

    @classmethod
    def search_repo(
            cls,
            search_query: str,
            db_name: str | None = None,
            *,
            names_only: bool = False,
            exact_match: bool = False,
    ) -> list[pyalpm.Package]:
        if "/" in search_query:
            db_name, search_query = search_query.split("/")
        result = []
        for sync_db in reversed(cls.get_alpm_handle().get_syncdbs()):
            if not db_name or db_name == sync_db.name:
                for pkg in sync_db.search(search_query):
                    if (
                            (
                                not names_only or search_query in pkg.name
                            ) and (
                                not exact_match or search_query in ([pkg.name, ] + pkg.groups)
                            )
                    ):
                        result.append(pkg)
        return list(reversed(result))

    @classmethod
    def get_repo_list(cls, *, quiet: bool = False) -> list[pyalpm.Package]:
        if not cls._packages_list_cache.get(PackageSource.REPO):
            with DbLockRepo():
                if not quiet:
                    print_stderr(translate("Reading repository package databases..."))
                cls._packages_list_cache[PackageSource.REPO] = cls.search_repo(
                    search_query=""
                )
        return cls._packages_list_cache[PackageSource.REPO]

    @classmethod
    def get_last_installed_package_date(cls) -> int:
        repo_names = []
        for repo in PackageDB.get_repo_list():
            repo_names.append(repo.name)
        packages = []
        for package in PackageDB.get_local_list():
            if package.name in repo_names:
                packages.append(package)
        packages_by_date = sorted(packages, key=lambda x: -x.installdate)
        return int(packages_by_date[0].installdate)

    @classmethod
    def get_print_format_output(
            cls, cmd_args: list[str], *, check_deps: bool = True, package_only: bool = False
    ) -> list[PacmanPrint]:
        cache_index = " ".join(sorted(cmd_args))
        cached_pkg = cls._pacman_find_cache.get(cache_index)
        if cached_pkg is not None:
            return cached_pkg
        results: list[PacmanPrint] = []
        final_args = cmd_args + ["--print-format", "%r/%n", ]
        if not check_deps and not package_only:
            final_args.append("--nodeps")
        if package_only:
            final_args += ["--nodeps", "--nodeps"]
        proc = spawn(final_args)
        if proc.returncode != 0:
            raise DependencyError((proc.stderr_text or "") + (proc.stdout_text or ""))
        found_packages_output = proc.stdout_text
        if found_packages_output:
            for line in found_packages_output.splitlines():
                try:
                    repo_name, pkg_name = line.split("/")
                except ValueError:
                    print_stderr(line)
                    continue
                else:
                    results.append(PacmanPrint(
                        full_name=line,
                        repo=repo_name,
                        name=pkg_name
                    ))
        cls._pacman_find_cache[cache_index] = results
        return results

    @classmethod
    def get_pacman_test_output(cls, cmd_args: list[str]) -> list[VersionMatcher]:
        if not cmd_args:
            return []
        cache_index = " ".join(sorted(cmd_args))
        cached_pkg = cls._pacman_test_cache.get(cache_index)
        if cached_pkg is not None:
            return cached_pkg
        results: list[VersionMatcher] = []
        not_found_packages_output = spawn(
            # pacman --deptest flag conflicts with some --sync options:
            get_pacman_command(ignore_args=[
                "overwrite"
            ]) + ["--deptest", ] + cmd_args
        ).stdout_text
        if not not_found_packages_output:
            cls._pacman_test_cache[cache_index] = results
            return results
        for line in not_found_packages_output.splitlines():
            try:
                version_matcher = VersionMatcher(line)
            except ValueError:
                print_stderr(line)
                continue
            else:
                results.append(version_matcher)
        return results

    @classmethod
    def get_not_found_repo_packages(cls, pkg_lines: list[str]) -> list[str]:

        pattern_notfound = create_pacman_pattern("target not found: %s\n")
        pattern_db_notfound = create_pacman_pattern("database not found: %s\n")

        not_found_pkg_names = []
        pkg_names_to_check: list[str] = []
        for pkg_name in pkg_lines:
            if pkg_name in cls._pacman_repo_pkg_present_cache:
                if not cls._pacman_repo_pkg_present_cache[pkg_name]:
                    not_found_pkg_names.append(pkg_name)
            else:
                pkg_names_to_check += pkg_name.split(",")

        if not pkg_names_to_check:
            return not_found_pkg_names

        results = (
            spawn(
                get_pacman_command() + ["--sync", "--print-format=%%", "--nodeps"] +
                pkg_names_to_check
            ).stderr_text or ""
        ).splitlines()
        new_not_found_pkg_names = []
        for result in results:
            for pattern in (pattern_notfound, pattern_db_notfound, ):
                groups = pattern.findall(result)
                if groups:
                    pkg_name = VersionMatcher(groups[0]).pkg_name
                    new_not_found_pkg_names.append(pkg_name)

        for pkg_name in pkg_names_to_check:
            cls._pacman_repo_pkg_present_cache[pkg_name] = pkg_name not in new_not_found_pkg_names

        return not_found_pkg_names + new_not_found_pkg_names

    @classmethod
    def get_not_found_local_packages(cls, pkg_lines: list[str]) -> list[str]:
        not_found_version_matchers = cls.get_pacman_test_output([
            splitted_pkg_name for splitted_pkg_names in
            [pkg_name.split(",") for pkg_name in pkg_lines]
            for splitted_pkg_name in splitted_pkg_names
        ])
        return list({
            vm.pkg_name for vm in not_found_version_matchers
        })

    @classmethod
    def find_repo_package(cls, pkg_name: str) -> pyalpm.Package:
        # @TODO: interactively ask for multiple providers and save the answer?
        if cls._pacman_repo_pkg_present_cache.get(pkg_name) is False:
            raise PackagesNotFoundInRepoError(packages=[pkg_name])
        all_repo_pkgs = PackageDB.get_repo_dict()
        try:
            results = cls.get_print_format_output(
                get_pacman_command() + ["--sync"] + pkg_name.split(","),
                package_only=True
            )
        except DependencyError as exc:
            raise PackagesNotFoundInRepoError(packages=[pkg_name]) from exc
        found_pkgs = [all_repo_pkgs[result.full_name] for result in results]
        if len(results) == 1:
            return found_pkgs[0]

        pkg_name = VersionMatcher(pkg_name).pkg_name
        for pkg in found_pkgs:
            if pkg.name == pkg_name:
                return pkg
        for pkg in found_pkgs:
            if pkg.provides:
                for provided_pkg_line in pkg.provides:
                    provided_name = VersionMatcher(provided_pkg_line).pkg_name
                    if provided_name == pkg_name:
                        return pkg
        raise PackagesNotFoundInRepoError(packages=[pkg_name])


def get_upgradeable_package_names() -> list[str]:
    upgradeable_packages_output = spawn(
        get_pacman_command() + ["--query", "--upgrades", "--quiet"]
    ).stdout_text
    if not upgradeable_packages_output:
        return []
    return upgradeable_packages_output.splitlines()


def find_upgradeable_packages() -> list[pyalpm.Package]:
    all_repo_pkgs = PackageDB.get_repo_dict()

    pkg_names = get_upgradeable_package_names()
    if not pkg_names:
        return []

    all_local_pkgs = PackageDB.get_local_dict()
    results: list[PacmanPrint] = []
    try:
        results = PackageDB.get_print_format_output(
            get_pacman_command() + ["--sync"] + pkg_names
        )
    except DependencyError as exc:
        print_error(translate("Dependencies can't be satisfied for the following packages:"))
        print_stderr(" " * 12 + " ".join(pkg_names))
        print_stderr(str(exc))
        for pkg_name in pkg_names:
            try:
                results += PackageDB.get_print_format_output(
                    get_pacman_command() + ["--sync"] + [pkg_name, ]
                )
            except DependencyError as exc2:
                print_error(translate("Because of:"))
                print_stderr(str(exc2))
    return [
        all_repo_pkgs[result.full_name] for result in results
        if result.name in all_local_pkgs
    ]


def find_sysupgrade_packages(ignore_pkgs: list[str] | None = None) -> list[pyalpm.Package]:
    all_repo_pkgs = PackageDB.get_repo_dict()

    extra_args: list[str] = []
    for excluded_pkg_name in ignore_pkgs or []:
        extra_args.append("--ignore")
        # pacman's --ignore doesn't work with repo name:
        extra_args.append(strip_repo_name(excluded_pkg_name))

    results = PackageDB.get_print_format_output(
        get_pacman_command() + ["--sync"] +
        ["--sysupgrade"] * parse_args().sysupgrade +
        extra_args
    )
    return [
        all_repo_pkgs[result.full_name] for result in results
    ]


def find_packages_not_from_repo() -> list[str]:
    local_pkg_names = PackageDB.get_local_pkgnames()
    repo_pkg_names = PackageDB.get_repo_pkgnames()
    not_found_packages = []
    for pkg_name in local_pkg_names:
        if pkg_name not in repo_pkg_names:
            not_found_packages.append(pkg_name)
    return not_found_packages


def refresh_pkg_db_if_needed() -> None:
    args = parse_args()
    if args.refresh:
        pacman_args = (sudo(
            get_pacman_command() + ["--sync"] + ["--refresh"] * args.refresh
        ))
        retry_interactive_command_or_exit(pacman_args)


def install_built_deps(
        deps_names_and_paths: dict[str, str],
        resolved_conflicts: list[list[str]] | None = None
) -> None:
    if not deps_names_and_paths:
        return

    local_packages = PackageDB.get_local_dict()
    args = parse_args()

    def _get_pacman_command() -> list[str]:
        return get_pacman_command() + reconstruct_args(args, ignore_args=[
            "upgrade",
            "asdeps",
            "sync",
            "sysupgrade",
            "refresh",
            "ignore",
            "downloadonly",
        ])

    explicitly_installed_deps = []
    for pkg_name, _path in deps_names_and_paths.items():
        _debug(pkg_name)
        if pkg_name in local_packages and local_packages[pkg_name].reason == 0:
            explicitly_installed_deps.append(pkg_name)
    deps_upgrade_success = True
    if len(explicitly_installed_deps) < len(deps_names_and_paths):
        deps_upgrade_success = retry_interactive_command(
            sudo(
                _get_pacman_command() + [
                    "--upgrade",
                    "--asdeps",
                ] + [
                    path for name, path in deps_names_and_paths.items()
                    if name not in explicitly_installed_deps
                ],
            ),
            pikspect=True,
            conflicts=resolved_conflicts,
        )

    explicit_upgrade_success = True
    if explicitly_installed_deps:
        explicit_upgrade_success = retry_interactive_command(
            sudo(
                _get_pacman_command() + [
                    "--upgrade",
                ] + [
                    path for name, path in deps_names_and_paths.items()
                    if name in explicitly_installed_deps
                ]
            ),
            pikspect=True,
            conflicts=resolved_conflicts,
        )

    PackageDB.discard_local_cache()

    if not (deps_upgrade_success and explicit_upgrade_success):
        raise DependencyError()


def strip_repo_name(pkg_name: str) -> str:
    return pkg_name.split("/", 1)[-1]


def get_ignored_pkgnames_from_patterns(
        orig_pkg_names: list[str],
        ignore_patterns: list[str]
) -> list[str]:
    ignored_pkg_names: list[str] = []
    for pkg_name in orig_pkg_names:
        for ignore_pattern in ignore_patterns:
            if fnmatch.fnmatch(pkg_name, ignore_pattern):
                ignored_pkg_names.append(pkg_name)
                break
    return ignored_pkg_names



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import shutil
import sys
import typing as t
from string import printable
from typing import Any, TextIO

from .args import parse_args
from .i18n import translate
from .lock import FancyLock

PADDING = 4


def color_enabled() -> bool:
    args = parse_args()
    if args.color == "never":
        return False
    if args.color == "always":
        return True
    try:
        if (sys.stderr.isatty() and sys.stdout.isatty()):
            return True
    except Exception:
        return False
    return False


class PrintLock(FancyLock):
    pass


def _print(
        destination: TextIO,
        message: Any = "",
        end: str = "\n",
        *,
        flush: bool = False,
        lock: bool = True
) -> None:
    # pylint: disable=unnecessary-dunder-call
    if not isinstance(message, str):
        message = str(message)
    if lock:
        PrintLock().__enter__()
    destination.write(f"{message}{end}")
    if flush:
        destination.flush()
    if lock:
        PrintLock().__exit__()


def print_stdout(
        message: Any = "",
        end: str = "\n",
        *,
        flush: bool = False,
        lock: bool = True
) -> None:
    _print(sys.stdout, message=message, end=end, flush=flush, lock=lock)


def print_stderr(
        message: Any = "",
        end: str = "\n",
        *,
        flush: bool = False,
        lock: bool = True
) -> None:
    _print(sys.stderr, message=message, end=end, flush=flush, lock=lock)


class Colors:
    black = 0
    red = 1
    green = 2
    yellow = 3
    blue = 4
    purple = 5
    cyan = 6
    white = 7


class ColorsHighlight:
    black = 8
    red = 9
    green = 10
    yellow = 11
    blue = 12
    purple = 13
    cyan = 14
    white = 15


def color_line(
        line: str, color_number: int, *, reset: bool = True, force: bool = False
) -> str:
    if not color_enabled() and not force:
        return line
    result = ""
    if color_number >= 8:
        result += "\033[0;1m"
        color_number -= 8
    result += f"\033[03{color_number}m{line}"
    # reset font:
    if reset:
        result += "\033[0;0m"
    return result


def bold_line(line: str) -> str:
    if not color_enabled():
        return line
    return f"\033[0;1m{line}\033[0m"


def print_warning(message: str = "") -> None:
    print_stderr(" ".join([
        color_line(":: " + translate("warning:"), ColorsHighlight.yellow),
        message
    ]))


def print_error(message: str) -> None:
    print_stderr(" ".join([
        color_line(":: " + translate("error:"), ColorsHighlight.red),
        message
    ]))


def print_debug(message: Any, *, lock: bool = True) -> None:
    args = parse_args()
    if not args.pikaur_debug:
        return
    prefix = translate("debug:")
    if args.debug:
        # to avoid mixing together with pacman's debug messages:
        prefix = translate("pikaur debug:")
    print_stderr(" ".join([
        color_line(":: " + prefix, Colors.cyan),
        str(message)
    ]), lock=lock)


def get_term_width() -> int:
    return shutil.get_terminal_size((80, 80)).columns


def format_paragraph(line: str) -> str:
    if not color_enabled():
        return PADDING * " " + line
    term_width = get_term_width()
    max_line_width = term_width - PADDING * 2

    result = []
    current_line: list[str] = []
    line_length = 0
    for word in line.split():
        if len(word) + line_length > max_line_width:
            result.append(current_line)
            current_line = []
            line_length = 0
        current_line.append(word)
        line_length += len(word) + 1
    result.append(current_line)

    return "\n".join([
        " ".join(
            [(PADDING - 1) * " ", ] +
            words +
            [(PADDING - 1) * " ", ],
        )
        for words in result
    ])


def range_printable(text: str, start: int = 0, end: int | None = None) -> str:
    if not end:
        end = len(text)

    result = ""
    counter = 0
    escape_seq = False
    for char in text:
        if counter >= start:
            result += char
        if not escape_seq and char in printable:
            counter += 1
        elif escape_seq and char == "m":
            escape_seq = False
        else:
            escape_seq = True
        if counter >= end:
            break
    return result


class DebugColorCounter:

    # cyan is purposely skipped as it's used in print_debug itself,
    # highlight-red is purposely skipped as it's used in print_error,
    # highlight-yellow is purposely skipped as it's used in print_warning:
    colors = [
        Colors.red,
        Colors.green,
        Colors.yellow,
        Colors.blue,
        Colors.purple,
        Colors.white,
        ColorsHighlight.green,
        ColorsHighlight.blue,
        ColorsHighlight.purple,
        ColorsHighlight.cyan,
        ColorsHighlight.white,
    ]
    _current_color_idx = 0

    @classmethod
    def get_next(cls) -> int:
        color = cls.colors[cls._current_color_idx]
        cls._current_color_idx += 1
        if cls._current_color_idx >= len(cls.colors):
            cls._current_color_idx = 0
        return color


def create_debug_logger(module_name: str, lock: bool | None = None) -> t.Callable[..., None]:
    color = DebugColorCounter.get_next()
    parent_lock = lock

    def debug(msg: Any, lock: bool | None = None) -> None:
        lock = lock if (lock is not None) else parent_lock
        msg = f"{color_line(module_name, color)}: {str(msg)}"
        if lock is not None:
            print_debug(msg, lock=lock)
        else:
            print_debug(msg)
    return debug



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import fcntl
import os
import pty
import re
import select
import shutil
import signal
import struct
import subprocess  # nosec B404
import sys
import termios
import tty
from multiprocessing.pool import ThreadPool
from time import sleep
from typing import Any, BinaryIO, Callable, TextIO

from .args import parse_args
from .core import DEFAULT_INPUT_ENCODING, get_sudo_refresh_command
from .pacman_i18n import _p
from .pprint import (
    ColorsHighlight,
    PrintLock,
    bold_line,
    color_line,
    create_debug_logger,
    get_term_width,
    print_stderr,
)

# SMALL_TIMEOUT = 0.1
SMALL_TIMEOUT = 0.01


TcAttrsType = list[int | list[bytes | int]]


_debug = create_debug_logger("pikspect", lock=False)


class TTYRestore():  # pragma: no cover

    old_tcattrs = None
    sub_tty_old_tcattrs = None

    @classmethod
    def save(cls) -> None:
        if sys.stdin.isatty():
            cls.old_tcattrs = termios.tcgetattr(sys.stdin.fileno())

    @classmethod
    def _restore(cls, what: TcAttrsType | None = None) -> None:
        # if sys.stdout.isatty():
        #     termios.tcdrain(sys.stdout.fileno())
        # if sys.stderr.isatty():
        #     termios.tcdrain(sys.stderr.fileno())
        # if sys.stdin.isatty():
        #     termios.tcflush(sys.stdin.fileno(), termios.TCIOFLUSH)
        if what:
            try:
                termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, what)
            except termios.error as exc:
                _debug(",".join(str(arg) for arg in exc.args))

    @classmethod
    def restore(cls, *_whatever: Any) -> None:
        cls._restore(cls.old_tcattrs)

    def __init__(self) -> None:
        try:
            self.sub_tty_old_tcattrs = termios.tcgetattr(sys.stdin.fileno())
        except termios.error:
            pass

    def restore_new(self, *_whatever: Any) -> None:
        self._restore(self.sub_tty_old_tcattrs)


TTYRestore.save()


def set_terminal_geometry(file_descriptor: int, rows: int, columns: int) -> None:
    term_geometry_struct = struct.pack("HHHH", rows, columns, 0, 0)
    fcntl.ioctl(
        file_descriptor, termios.TIOCSWINSZ, term_geometry_struct
    )


class TTYInputWrapper():  # pragma: no cover

    tty_opened = False

    def __init__(self) -> None:
        self.is_pipe = not sys.stdin.isatty()

    def __enter__(self) -> None:
        if self.is_pipe:
            self.old_stdin = sys.stdin
            try:
                _debug("Attaching to TTY manually...")
                sys.stdin = open("/dev/tty", encoding=DEFAULT_INPUT_ENCODING)
                self.tty_opened = True
            except Exception as exc:
                _debug(exc)

    def __exit__(self, *_exc_details: Any) -> None:
        if self.is_pipe and self.tty_opened:
            _debug("Restoring stdin...", lock=False)
            sys.stdin.close()
            sys.stdin = self.old_stdin


class NestedTerminal():

    def __init__(self) -> None:
        self.tty_wrapper = TTYInputWrapper()

    def __enter__(self) -> os.terminal_size:
        _debug("Opening virtual terminal...")
        self.tty_wrapper.__enter__()
        real_term_geometry = shutil.get_terminal_size((80, 80))
        for stream in (
                sys.stdin,
                sys.stderr,
                sys.stdout,
        ):
            if stream.isatty():
                tty.setcbreak(stream.fileno())
        return real_term_geometry

    def __exit__(self, *exc_details: Any) -> None:
        self.tty_wrapper.__exit__(*exc_details)
        TTYRestore.restore()


def _match(pattern: str, line: str) -> bool:
    return len(line) >= len(pattern) and bool(
        re.compile(pattern).search(line)
        if ".*" in pattern else
        (pattern in line)
    )


class PikspectSignalHandler():

    signal_handler: Callable[..., Any] | None = None

    @classmethod
    def set_handler(cls, signal_handler: Callable[..., Any]) -> None:
        cls.signal_handler = signal_handler

    @classmethod
    def clear(cls) -> None:
        cls.signal_handler = None

    @classmethod
    def get(cls) -> Callable[..., Any] | None:
        return cls.signal_handler


class PikspectPopen(subprocess.Popen[bytes]):

    print_output: bool
    capture_input: bool
    capture_output: bool
    historic_output: list[bytes]
    pty_in: TextIO
    pty_out: BinaryIO
    default_questions: dict[str, list[str]]
    # max_question_length = 0  # preserve enough information to analyze questions
    max_question_length = get_term_width() * 2  # preserve also at least last line
    # write buffer:
    _write_buffer: bytes = b""
    # some help for mypy:
    output: bytes = b""

    def __init__(
            self,
            args: list[str],
            *,
            print_output: bool = True,
            capture_input: bool = True,
            capture_output: bool = False,
            default_questions: dict[str, list[str]] | None = None
    ) -> None:
        self.args = args
        self.print_output = print_output
        self.capture_input = capture_input
        self.capture_output = capture_output
        self.default_questions = {}
        self.historic_output = []
        if default_questions:
            self.add_answers(default_questions)

        self.pty_user_master, self.pty_user_slave = pty.openpty()
        self.pty_cmd_master, self.pty_cmd_slave = pty.openpty()

        super().__init__(
            args=args,
            stdin=self.pty_user_slave,
            stdout=self.pty_cmd_slave,
            stderr=self.pty_cmd_slave,
        )

    def __del__(self) -> None:
        self.terminate()
        self.communicate()

    def add_answers(self, extra_questions: dict[str, list[str]]) -> None:
        for answer, questions in extra_questions.items():
            self.default_questions[answer] = self.default_questions.get(answer, []) + questions
            for question in questions:
                if len(question) > self.max_question_length:
                    self.max_question_length = len(question.encode(DEFAULT_INPUT_ENCODING))
        self.check_questions()

    def communicator_thread(self) -> int:
        result: int = self._wait(None)  # type: ignore[attr-defined]
        return result

    def run(self) -> None:
        if not isinstance(self.args, list):
            raise TypeError("`args` should be list")
        PikspectSignalHandler.set_handler(
            lambda *_whatever: self.send_signal(signal.SIGINT)
        )
        try:
            with NestedTerminal() as real_term_geometry:

                if "sudo" in self.args:  # pragma: no cover
                    subprocess.run(  # nosec B603
                        get_sudo_refresh_command(),
                        check=True
                    )
                with open(
                        self.pty_user_master, "w", encoding=DEFAULT_INPUT_ENCODING
                ) as self.pty_in:
                    with open(self.pty_cmd_master, "rb", buffering=0) as self.pty_out:
                        set_terminal_geometry(
                            self.pty_out.fileno(),
                            columns=real_term_geometry.columns,
                            rows=real_term_geometry.lines
                        )
                        with ThreadPool(processes=3) as pool:
                            output_task = pool.apply_async(self.cmd_output_reader_thread, ())
                            input_task = pool.apply_async(self.user_input_reader_thread, ())
                            communicate_task = pool.apply_async(self.communicator_thread, ())
                            pool.close()

                            output_task.get()
                            sys.stdout.buffer.write(self._write_buffer)
                            sys.stdout.buffer.flush()
                            communicate_task.get()
                            input_task.get()
                            pool.join()
        finally:
            PikspectSignalHandler.clear()
            os.close(self.pty_cmd_slave)
            os.close(self.pty_user_slave)

    def check_questions(self) -> None:
        try:
            historic_output = b"".join(self.historic_output).decode(DEFAULT_INPUT_ENCODING)
        except UnicodeDecodeError:  # pragma: no cover
            return

        clear_buffer = False

        for answer, questions in self.default_questions.items():
            for question in questions:
                if not _match(question, historic_output):
                    continue
                self.write_something((answer + "\n").encode(DEFAULT_INPUT_ENCODING))
                with PrintLock():
                    self.pty_in.write(answer)
                    sleep(SMALL_TIMEOUT)
                    self.pty_in.write("\n")
                    self.pty_in.flush()
                clear_buffer = True
                break

        if clear_buffer:
            self.historic_output = [b""]

    def write_buffer_contents(self) -> None:
        if self._write_buffer:
            sys.stdout.buffer.write(self._write_buffer)
            sys.stdout.buffer.flush()
            self._write_buffer = b""

    def write_something(self, output: bytes) -> None:
        if not (self.print_output or self.capture_output):
            return
        with PrintLock():
            if self.capture_output:
                self.output += output
            self._write_buffer += output
            self.write_buffer_contents()

    def cmd_output_reader_thread(self) -> None:
        while True:

            try:
                selected = select.select([self.pty_out, ], [], [], SMALL_TIMEOUT)
            except ValueError:  # pragma: no cover
                return
            else:
                readers = selected[0]
            if not readers:
                if self.returncode is not None:
                    break
                if not self.historic_output:
                    sleep(SMALL_TIMEOUT)
                self.write_buffer_contents()
                continue
            pty_reader = readers[0]
            output = pty_reader.read(4096)

            self.historic_output = (
                self.historic_output[-self.max_question_length:] + [output, ]
            )
            self.write_something(output)
            self.check_questions()

    def user_input_reader_thread(self) -> None:  # pragma: no cover
        while self.returncode is None:
            if not self.capture_input:
                sleep(SMALL_TIMEOUT)
                continue

            char = None

            if sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
                char = sys.stdin.read(1)
                if char in [None, ""]:
                    sleep(SMALL_TIMEOUT)
                    continue
            else:
                sleep(SMALL_TIMEOUT)
                continue

            try:
                with PrintLock():
                    if ord(char) == 127:  # BackSpace
                        sys.stdout.write("\b \b")
                    if ord(char) == 23:  # Ctrl+W
                        sys.stdout.write(
                            "\r" + " " * get_term_width() + "\r" +
                            b"".join(
                                self.historic_output
                            ).decode(DEFAULT_INPUT_ENCODING).splitlines()[-1]
                        )
                    self.pty_in.write(char)
                    self.pty_in.flush()
            except ValueError as exc:
                print(exc)  # noqa: T201
            self.write_something(char.encode(DEFAULT_INPUT_ENCODING))


class YesNo:
    ANSWER_Y = _p("Y")
    ANSWER_N = _p("N")
    QUESTION_YN_YES = _p("[Y/n]")
    QUESTION_YN_NO = _p("[y/N]")


def format_pacman_question(message: str, question: str = YesNo.QUESTION_YN_YES) -> str:
    return bold_line(f" {_p(message)} {question} ")


def pikspect(
        cmd: list[str],
        *,
        print_output: bool = True,
        auto_proceed: bool = True,
        conflicts: list[list[str]] | None = None,
        extra_questions: dict[str, list[str]] | None = None,
        capture_output: bool | None = None,
) -> PikspectPopen:

    class Questions:
        PROCEED = [
            format_pacman_question("Proceed with installation?"),
            format_pacman_question("Proceed with download?"),
        ]
        REMOVE = [
            format_pacman_question("Do you want to remove these packages?"),
        ]
        CONFLICT = format_pacman_question(
            "%s and %s are in conflict. Remove %s?", YesNo.QUESTION_YN_NO
        )
        CONFLICT_VIA_PROVIDED = format_pacman_question(
            "%s and %s are in conflict (%s). Remove %s?", YesNo.QUESTION_YN_NO
        )

    def format_conflicts(conflicts: list[list[str]]) -> list[str]:
        return [
            Questions.CONFLICT % (new_pkg, old_pkg, old_pkg)
            for new_pkg, old_pkg in conflicts
        ] + [
            (
                re.escape(Questions.CONFLICT_VIA_PROVIDED % (
                    new_pkg, old_pkg, ".*", old_pkg
                ))
            ).replace(r"\.\*", ".*")
            for new_pkg, old_pkg in conflicts
        ]

    default_questions: dict[str, list[str]] = {}
    if auto_proceed:
        default_questions = {
            YesNo.ANSWER_Y: Questions.PROCEED + Questions.REMOVE,
            YesNo.ANSWER_N: [],
        }

    with PikspectPopen(
        cmd,
        print_output=print_output,
        default_questions=default_questions,
        capture_output=bool(capture_output),
    ) as proc:

        extra_questions = extra_questions or {}
        if conflicts:
            extra_questions[YesNo.ANSWER_Y] = (
                extra_questions.get(YesNo.ANSWER_Y, []) + format_conflicts(conflicts)
            )
        if extra_questions:
            proc.add_answers(extra_questions)

        if parse_args().print_commands:
            print_stderr(
                color_line("pikspect => ", ColorsHighlight.cyan) +
                " ".join(cmd)
            )
        proc.run()
        return proc


if __name__ == "__main__":
    pikspect(sys.argv[1:])



"""Licensed under GPLv3, see https://www.gnu.org/licenses/ ."""

from itertools import zip_longest
from typing import Callable

import pyalpm

VERSION_SEPARATORS = (".", "+", "-", ":")


def compare_versions(version1: str, version2: str) -> int:
    """
    `vercmp` is used to determine the relationship between two given version numbers.
    It outputs values as follows:
        < 0 : if ver1 < ver2
        = 0 : if ver1 == ver2
        > 0 : if ver1 > ver2
    """
    return pyalpm.vercmp(version1, version2)


class VersionMatcher():
    """
    Represents version string as stated in dependencies (e.g. `>=1.23`).
    And can match that pattern against some version.
    """

    version: str | None = None
    version_matchers: list[Callable[[str], int]]
    line: str
    pkg_name: str

    def __call__(self, version: str | None) -> int:
        """Check if `version` matching our `VersionMatcher`."""
        if not version:
            return True
        return min(
            version_matcher(version)
            for version_matcher in self.version_matchers
        )

    def __init__(self, depend_line: str, *, is_pkg_deps: bool = False) -> None:
        self.line = depend_line
        self._set_version_matcher_func(is_pkg_deps=is_pkg_deps)

    def add_version_matcher(self, version_matcher: "VersionMatcher") -> None:
        """
        Embed another `VersionMatcher` into this.
        So they both could check against some version
        when this `VersionMatcher` is called.
        """
        if version_matcher.line in self.line.split(","):
            return
        self.version_matchers.append(version_matcher.version_matchers[0])
        self.line += "," + version_matcher.line
        if self.version:
            if version_matcher.version:
                self.version += "," + version_matcher.version
        else:
            self.version = version_matcher.version

    def _set_version_matcher_func(  # pylint: disable=too-many-locals
            self, *, is_pkg_deps: bool = False
    ) -> None:
        # pylint: disable=invalid-name

        version: str | None = None

        def get_version() -> str | None:
            return version

        def cmp_lt(v: str) -> int:
            self_version = get_version()
            if not self_version:
                return cmp_default(v)
            return compare_versions(v, self_version) < 0

        def cmp_gt(v: str) -> int:
            self_version = get_version()
            if not self_version:
                return cmp_default(v)
            return compare_versions(v, self_version) > 0

        def cmp_eq(v: str) -> int:
            self_version = get_version()
            if not self_version:
                return cmp_default(v)
            if is_pkg_deps:
                common_version, _ = get_common_version(v, self_version)
                if common_version == self_version:
                    return True
            return compare_versions(v, self_version) == 0

        def cmp_le(v: str) -> int:
            return cmp_eq(v) or cmp_lt(v)

        def cmp_ge(v: str) -> int:
            return cmp_eq(v) or cmp_gt(v)

        def cmp_default(v: str) -> int:  # noqa: ARG001  # pylint: disable=unused-argument
            return 1

        cond: str | None = None
        version_matcher = cmp_default
        for test_cond, matcher in (
                (">=", cmp_ge),
                ("<=", cmp_le),
                ("=", cmp_eq),
                (">", cmp_gt),
                ("<", cmp_lt),
        ):
            if test_cond in self.line:
                cond = test_cond
                version_matcher = matcher
                break

        if cond:
            splitted_line = self.line.split(cond)
            pkg_name = splitted_line[0]
            version = cond.join(splitted_line[1:])
        else:
            pkg_name = self.line

        self.pkg_name = pkg_name
        self.version = version
        self.version_matchers = [version_matcher]

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"{self.pkg_name}{[m.__name__ for m in self.version_matchers]}{self.version}>"
        )


def split_version(version: str) -> list[str]:
    """Split version, e.g. `"1.2+3"` to `["1", "2", "3"]`."""
    splitted_version = []
    block = ""
    for char in version:
        if char in VERSION_SEPARATORS:
            splitted_version.append(block)
            splitted_version.append(char)
            block = ""
        else:
            block += char
    if block != "":
        splitted_version.append(block)
    return splitted_version


def split_always(line: str, separator: str, *, pad_right: bool = False) -> tuple[str, str]:
    """
    Same as builtin string `.split()`,
    but return empty string on the left or right if no separator found.
    """
    splitted_line = line.split(separator, 1)
    if len(splitted_line) > 1:
        return splitted_line[0], separator + splitted_line[1]
    if pad_right:
        return splitted_line[0], ""
    return "", splitted_line[0]


def rsplit_always(line: str, separator: str) -> tuple[str, str]:
    """
    Same as builtin string `.rsplit()`,
    but return empty string on right if no separator found.
    """
    splitted_line = line.rsplit(separator, 1)
    if len(splitted_line) > 1:
        return splitted_line[0] + separator, splitted_line[1]
    return splitted_line[0], ""


def get_common_version(version1: str, version2: str) -> tuple[str, int]:
    """
    Get common part of two versions and compute their difference "weight".
    For example if epoch changes it affects weight by 1000,
    major version by 500 and etc.
    E.g. for `1.2.3` and `1.2.5` this would be `("1.2.", 96)`,
    for `1.2.3` and `2.0.0` - `("", 500)`.
    """

    def _split_epoch(version: str) -> tuple[str, str]:
        return split_always(version, ":")

    def _split_major(version: str) -> tuple[str, str]:
        return split_always(version, ".", pad_right=True)

    def _split_release(version: str) -> tuple[str, str]:
        return rsplit_always(version, "-")

    common_string = ""
    diff_weight = 0
    diff_found = False
    if "" in (version1, version2):
        return common_string, diff_weight
    for weight, version_chunk1, version_chunk2 in (
            (
                1000, _split_epoch(version1)[0], _split_epoch(version2)[0]
            ),
            (
                500, _split_release(_split_major(_split_epoch(version1)[1])[0])[0],
                _split_release(_split_major(_split_epoch(version2)[1])[0])[0],
            ),
            (
                100, _split_release(_split_major(_split_epoch(version1)[1])[1])[0],
                _split_release(_split_major(_split_epoch(version2)[1])[1])[0],
            ),
            (
                10, _split_release(version1)[1], _split_release(version2)[1],
            ),
    ):
        for block1, block2 in zip_longest(
                split_version(version_chunk1),
                split_version(version_chunk2),
                fillvalue=" "
        ):
            if block1 == block2:
                if not diff_found:
                    common_string += block1
            else:
                diff_found = True
                if diff_weight == 0 and block1 not in VERSION_SEPARATORS:
                    diff_weight += weight
            weight -= 1
    if version2 == "devel":
        diff_weight = 9999
    return common_string, diff_weight


def get_version_diff(version: str, common_version: str) -> str:
    """
    Get the different part of the version and version prefix returned by `get_common_version()`.
    E.g. if version is '1.2.3' and common part - '1.2.' - the different part would be '3'.
    """
    if common_version == "":
        return version
    return common_version.join(
        version.split(common_version)[1:]
    )



"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import sys
from datetime import datetime
from fnmatch import fnmatch
from typing import TYPE_CHECKING, Iterable, Sequence, TypeVar

import pyalpm

from .args import parse_args
from .aur import AURPackageInfo
from .config import VERSION, PikaurConfig
from .core import AURInstallInfo, InstallInfo, PackageSource, RepoInstallInfo
from .i18n import translate, translate_many
from .pacman import OFFICIAL_REPOS, PackageDB
from .pprint import (
    Colors,
    ColorsHighlight,
    bold_line,
    color_line,
    format_paragraph,
    get_term_width,
    print_stderr,
    print_stdout,
    print_warning,
)
from .version import get_common_version, get_version_diff

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .install_info_fetcher import InstallInfoFetcher


GROUP_COLOR = Colors.blue
REPLACEMENTS_COLOR = ColorsHighlight.cyan
ORPHANED_COLOR = ColorsHighlight.red


AnyPackage = AURPackageInfo | pyalpm.Package
InstallInfoT = TypeVar("InstallInfoT", bound=InstallInfo)


def print_version(pacman_version: str, pyalpm_version: str, *, quiet: bool = False) -> None:
    if quiet:
        print_stdout(f"Pikaur v{VERSION}")
        print_stdout(f"{pacman_version} - pyalpm v{pyalpm_version}")
    else:
        year = str(datetime.now().year)
        sys.stdout.write(r"""
      /:}               _
     /--1             / :}
    /   |           / `-/
   |  ,  --------  /   /     Pikaur v""" + VERSION + r"""
   |'                 Y      (C) 2018-""" + year + r""" Pikaur development team
  /                   l      Licensed under GPLv3
  l  /       \        l
  j  ●   .   ●        l      """ + pacman_version + r"""
 { )  ._,.__,   , -.  {      pyalpm v""" + pyalpm_version + r"""
  Y    \  _/     ._/   \

""")


def print_not_found_packages(not_found_packages: list[str], *, repo: bool = False) -> None:
    num_packages = len(not_found_packages)
    print_warning(
        bold_line(
            translate_many(
                "Following package cannot be found in repositories:",
                "Following packages cannot be found in repositories:",
                num_packages
            )
            if repo else
            translate_many(
                "Following package cannot be found in AUR:",
                "Following packages cannot be found in AUR:",
                num_packages
            )
        )
    )
    for package in not_found_packages:
        print_stderr(format_paragraph(package))


class RepoColorGenerator:

    _type_storage: dict[str, int] = {}
    _cache: dict[str, dict[str, int]] = {}
    _init_done = False

    @classmethod
    def do_init(cls) -> None:
        if cls._init_done:
            return

        cls._init_done = True
        for official_repo_name in OFFICIAL_REPOS:
            cls.get_next("repo", official_repo_name)
        for repo in PackageDB.get_alpm_handle().get_syncdbs():
            cls.get_next("repo", repo.name)

    @classmethod
    def get_next(cls, color_type: str, _id: str) -> int:
        cls.do_init()
        if cls._cache.get(color_type, {}).get(_id):
            return cls._cache[color_type][_id]

        if (
                not cls._type_storage.get(color_type) or
                cls._type_storage[color_type] >= 15
        ):
            cls._type_storage[color_type] = 10
        else:
            cls._type_storage[color_type] += 1

        cls._cache.setdefault(color_type, {})
        cls._cache[color_type][_id] = cls._type_storage[color_type]
        return cls._cache[color_type][_id]


def pretty_format_repo_name(repo_name: str, *, color: bool = True) -> str:
    result = f"{repo_name}/"
    if not color:
        return result
    return color_line(result, RepoColorGenerator.get_next("repo", repo_name))


def pretty_format_upgradeable(  # pylint: disable=too-many-statements
        packages_updates: Sequence[InstallInfo],
        *,
        verbose: bool = False,
        print_repo: bool = False,
        color: bool = True,
        template: str | None = None
) -> str:
    parent_color = color

    def _color_line(line: str, color: int, reset: bool | None = None) -> str:
        if not parent_color:
            return line
        if reset is not None:
            return color_line(line, color_number=color, reset=reset)
        return color_line(line, color_number=color)

    def _bold_line(line: str) -> str:
        return bold_line(line) if parent_color else line

    # @TODO: python 3.11
    # type_sort_key = tuple[str, *tuple[int, ...]] | str
    type_sort_key = (
        tuple[str, str] | tuple[str, int] |  # type: ignore[misc]
        tuple[int, str] | str  # type: ignore[misc]
    )

    def pretty_format(  # pylint:disable=too-many-locals,R0912
            pkg_update: "InstallInfo"
    ) -> tuple[str, type_sort_key]:
        common_version, diff_weight = get_common_version(
            pkg_update.current_version or "", pkg_update.new_version or ""
        )
        user_config = PikaurConfig()
        color_config = user_config.colors
        version_color = color_config.Version.get_int()
        old_color = color_config.VersionDiffOld.get_int()
        new_color = color_config.VersionDiffNew.get_int()
        column_width = min(int(get_term_width() / 2.5), 37)

        sort_by: type_sort_key = (
            -diff_weight,
            pkg_update.name
        )
        user_chosen_sorting = user_config.sync.UpgradeSorting
        if user_chosen_sorting == "pkgname":
            sort_by = pkg_update.name
        elif user_chosen_sorting == "repo":
            sort_by = (
                pkg_update.repository or "zzz_(aur)",
                pkg_update.name
            )

        pkg_name = pkg_update.name
        pkg_len = len(pkg_update.name)

        pkg_name = _bold_line(pkg_name)
        if (print_repo or verbose) and pkg_update.repository:
            pkg_name = "{}{}".format(  # pylint: disable=consider-using-f-string
                pretty_format_repo_name(pkg_update.repository, color=color),
                pkg_name
            )
            pkg_len += len(pkg_update.repository) + 1
        elif print_repo:
            pkg_name = "{}{}".format(  # pylint: disable=consider-using-f-string
                _color_line("aur/", ColorsHighlight.red),
                pkg_name
            )
            pkg_len += len("aur/")

        if pkg_update.required_by:
            required_by = " ({})".format(  # pylint: disable=consider-using-f-string
                translate("for {pkg}").format(
                    pkg=", ".join([p.package.name for p in pkg_update.required_by])
                )
            )
            pkg_len += len(required_by)
            dep_color = Colors.yellow
            required_by = _color_line(" ({})", dep_color).format(
                translate("for {pkg}").format(
                    pkg=_color_line(", ", dep_color).join([
                        _color_line(p.package.name, dep_color + 8) for p in pkg_update.required_by
                    ]) + _color_line("", dep_color, reset=False),
                )
            )
            pkg_name += required_by
        if pkg_update.provided_by:
            provided_by = " ({})".format(  # pylint: disable=consider-using-f-string
                " # ".join([p.name for p in pkg_update.provided_by])
            )
            pkg_len += len(provided_by)
            pkg_name += _color_line(provided_by, Colors.green)
        if pkg_update.members_of:
            members_of = " ({})".format(  # pylint: disable=consider-using-f-string
                translate_many("{grp} group", "{grp} groups", len(pkg_update.members_of)).format(
                    grp=", ".join(g for g in pkg_update.members_of),
                )
            )
            pkg_len += len(members_of)
            members_of = _color_line(" ({})", GROUP_COLOR).format(
                translate_many("{grp} group", "{grp} groups", len(pkg_update.members_of)).format(
                    grp=_color_line(", ", GROUP_COLOR).join(
                        [_color_line(g, GROUP_COLOR + 8) for g in pkg_update.members_of]
                    ) + _color_line("", GROUP_COLOR, reset=False),
                )
            )
            pkg_name += _color_line(members_of, GROUP_COLOR)
        if pkg_update.replaces:
            replaces = " (replaces {})".format(  # pylint: disable=consider-using-f-string
                ", ".join(g for g in pkg_update.replaces)
            )
            pkg_len += len(replaces)
            pkg_name += _color_line(replaces, REPLACEMENTS_COLOR)
            if not color:
                pkg_name = f"# {pkg_name}"

        pkg_size = ""
        if (
                user_config.sync.ShowDownloadSize.get_bool() and
                isinstance(pkg_update.package, pyalpm.Package)
        ):
            pkg_size = f"{pkg_update.package.size/1024/1024:.2f} MiB"

        days_old = ""
        if pkg_update.devel_pkg_age_days:
            days_old = " " + translate("({} days old)").format(pkg_update.devel_pkg_age_days)

        if (
                isinstance(pkg_update.package, AURPackageInfo) and
                pkg_update.maintainer is None
        ):
            orphaned = f" [{translate('orphaned')}]"
            pkg_len += len(orphaned)
            pkg_name += _color_line(orphaned, ORPHANED_COLOR)

        out_of_date = ""
        if (
                isinstance(pkg_update.package, AURPackageInfo) and
                pkg_update.package.outofdate is not None
        ):
            out_of_date = _color_line(
                " [{}: {}]".format(  # pylint: disable=consider-using-f-string
                    translate("outofdate"),
                    datetime.fromtimestamp(pkg_update.package.outofdate).strftime("%Y/%m/%d")
                ),
                color_config.VersionDiffOld.get_int()
            )

        return (
            template or (
                " {pkg_name}{spacing}"
                " {current_version}{spacing2}"
                "{version_separator}{new_version}{spacing3}"
                "{pkg_size}{days_old}{out_of_date}{verbose}"
            )
        ).format(
            pkg_name=pkg_name,
            current_version=(
                _color_line(common_version, version_color) +
                _color_line(
                    get_version_diff(pkg_update.current_version or "", common_version),
                    old_color
                )
            ),
            new_version=(
                _color_line(common_version, version_color) +
                _color_line(
                    get_version_diff(pkg_update.new_version or "", common_version),
                    new_color
                )
            ),
            version_separator=(
                " -> " if (pkg_update.current_version or pkg_update.new_version) else ""
            ),
            spacing=" " * max(1, (column_width - pkg_len)),
            spacing2=" " * max(1, (
                column_width - 18 -
                len(pkg_update.current_version or "") -
                max(-1, (pkg_len - column_width))
            )),
            spacing3=(" " * max(1, (
                column_width - 18 -
                len(pkg_update.new_version or "") -
                max(-1, (pkg_len - column_width))
            )) if pkg_size else ""),
            pkg_size=pkg_size,
            days_old=days_old,
            out_of_date=out_of_date,
            verbose=(
                "" if not (verbose and pkg_update.description)
                else f"\n{format_paragraph(pkg_update.description)}"
            )
        ), sort_by

    return "\n".join([
        line for line, _ in sorted(
            [
                pretty_format(pkg_update)
                for pkg_update in packages_updates
            ],
            key=lambda x: x[1],
        )
    ])


class SysupgradePrettyFormatter:

    def __init__(
        self,
        install_info: "InstallInfoFetcher",
        *,
        verbose: bool,
        manual_package_selection: bool
    ) -> None:
        self.color = True
        self.install_info = install_info
        self.verbose = verbose

        self.repo_packages_updates: list[RepoInstallInfo] = \
            install_info.repo_packages_install_info[::]
        self.thirdparty_repo_packages_updates: list[RepoInstallInfo] = \
            install_info.thirdparty_repo_packages_install_info[::]
        self.aur_updates: list[AURInstallInfo] = \
            install_info.aur_updates_install_info[::]
        self.repo_replacements: list[RepoInstallInfo] = \
            install_info.repo_replacements_install_info[::]
        self.thirdparty_repo_replacements: list[RepoInstallInfo] = \
            install_info.thirdparty_repo_replacements_install_info[::]

        self.new_repo_deps: list[RepoInstallInfo] = \
            install_info.new_repo_deps_install_info[::]
        self.new_thirdparty_repo_deps: list[RepoInstallInfo] = \
            install_info.new_thirdparty_repo_deps_install_info[::]
        self.new_aur_deps: list[AURInstallInfo] = \
            install_info.aur_deps_install_info[::]

        if manual_package_selection:
            self.color = False
            self.new_repo_deps = []
            self.new_thirdparty_repo_deps = []
            self.new_aur_deps = []

        self.all_install_info_lists: Sequence[
            list[AURInstallInfo] | list[RepoInstallInfo]
        ] = [
            self.repo_packages_updates,
            self.thirdparty_repo_packages_updates,
            self.aur_updates,
            self.repo_replacements,
            self.thirdparty_repo_replacements,
            self.new_repo_deps,
            self.new_thirdparty_repo_deps,
            self.new_aur_deps,
        ]

        self.config = PikaurConfig()
        self.result: list[str] = []

    def _color_line(self, line: str, color: int) -> str:
        return color_line(line, color_number=color) if self.color else line

    def _bold_line(self, line: str) -> str:
        return bold_line(line) if self.color else line

    def pretty_format_upgradeable(
            self,
            install_infos: Sequence[InstallInfo],
            print_repo: bool | None = None
    ) -> str:
        if print_repo is None:
            print_repo = self.config.sync.AlwaysShowPkgOrigin.get_bool()
        return pretty_format_upgradeable(
            install_infos,
            verbose=self.verbose, color=self.color, print_repo=print_repo
        )

    def pformat_warned_packages(self) -> None:
        warn_about_packages_str = self.config.ui.WarnAboutPackageUpdates.get_str()
        warn_about_packages_list: list[InstallInfo] = []

        def remove_globs_from_pkg_list(pkg_list: list[InstallInfoT]) -> None:
            pkg_install_info: InstallInfoT
            for pkg_install_info in pkg_list[::]:
                for glob in globs_and_names:
                    if fnmatch(pkg_install_info.name, glob):
                        pkg_list.remove(pkg_install_info)
                        warn_about_packages_list.append(pkg_install_info)

        if warn_about_packages_str:
            globs_and_names = warn_about_packages_str.split(",")
            pkg_list: list[RepoInstallInfo] | list[AURInstallInfo]
            for pkg_list in self.all_install_info_lists:
                remove_globs_from_pkg_list(pkg_list)  # type: ignore[misc]

        if warn_about_packages_list:
            self.result.append("\n{} {} {} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.blue),
                self._color_line("!!", ColorsHighlight.red),
                self._color_line(
                    translate_many(
                        "WARNING about package installation:",
                        "WARNING about packages installation:",
                        len(warn_about_packages_list)
                    ), ColorsHighlight.red
                ),
                self._color_line("!!", ColorsHighlight.red),
            ))
            self.result.append(self.pretty_format_upgradeable(warn_about_packages_list))

    def pformat_replacements(self) -> None:
        if self.repo_replacements:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.blue),
                self._bold_line(translate_many(
                    "Repository package suggested as a replacement:",
                    "Repository packages suggested as a replacement:",
                    len(self.repo_replacements)))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.repo_replacements,
            ))
        if self.thirdparty_repo_replacements:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.blue),
                self._bold_line(translate_many(
                    "Third-party repository package suggested as a replacement:",
                    "Third-party repository packages suggested as a replacement:",
                    len(self.repo_packages_updates)))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.thirdparty_repo_replacements,
            ))

    def pformat_repo(self) -> None:
        if self.repo_packages_updates:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.blue),
                self._bold_line(translate_many(
                    "Repository package will be installed:",
                    "Repository packages will be installed:",
                    len(self.repo_packages_updates)))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.repo_packages_updates,
            ))
        if self.new_repo_deps:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.yellow),
                self._bold_line(translate_many(
                    "New dependency will be installed from repository:",
                    "New dependencies will be installed from repository:",
                    len(self.new_repo_deps)
                ))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.new_repo_deps,
            ))

    def pformat_thirdaprty_repo(self) -> None:
        if self.thirdparty_repo_packages_updates:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.blue),
                self._bold_line(translate_many(
                    "Third-party repository package will be installed:",
                    "Third-party repository packages will be installed:",
                    len(self.thirdparty_repo_packages_updates)
                ))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.thirdparty_repo_packages_updates,
                print_repo=True
            ))
        if self.new_thirdparty_repo_deps:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.yellow),
                self._bold_line(translate_many(
                    "New dependency will be installed from third-party repository:",
                    "New dependencies will be installed from third-party repository:",
                    len(self.new_thirdparty_repo_deps)
                ))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.new_thirdparty_repo_deps,
            ))

    def pformat_aur(self) -> None:
        if self.aur_updates:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.cyan),
                self._bold_line(translate_many(
                    "AUR package will be installed:",
                    "AUR packages will be installed:",
                    len(self.aur_updates)
                ))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.aur_updates,
                print_repo=False
            ))
        if self.new_aur_deps:
            self.result.append("\n{} {}".format(  # pylint: disable=consider-using-f-string
                self._color_line("::", ColorsHighlight.yellow),
                self._bold_line(translate_many(
                    "New dependency will be installed from AUR:",
                    "New dependencies will be installed from AUR:",
                    len(self.new_aur_deps)
                ))
            ))
            self.result.append(self.pretty_format_upgradeable(
                self.new_aur_deps,
                print_repo=False
            ))

    def pformat_total_size(self) -> None:
        if self.config.sync.ShowDownloadSize.get_bool():
            self.result.append(
                "\n" +
                self._bold_line(translate("Total Download Size:")) +
                f"{str(round(self.install_info.get_total_download_size(), 2)).rjust(10)} MiB"
                "\n" +
                self._bold_line(translate("Total Installed Size:")) +
                f"{str(round(self.install_info.get_total_installed_size(), 2)).rjust(9)} MiB"
                "\n"
            )
        else:
            self.result += [""]

    def __call__(self) -> str:
        self.pformat_warned_packages()
        self.pformat_replacements()
        self.pformat_repo()
        self.pformat_thirdaprty_repo()
        self.pformat_aur()
        self.pformat_total_size()
        return "\n".join(self.result)


def pretty_format_sysupgrade(
        install_info: "InstallInfoFetcher",
        *,
        verbose: bool = False,
        manual_package_selection: bool = False
) -> str:
    return SysupgradePrettyFormatter(
        install_info=install_info,
        verbose=verbose,
        manual_package_selection=manual_package_selection
    )()


def print_ignored_package(
        package_name: str | None = None,
        install_info: InstallInfo | None = None,
        ignored_from: str | None = None
) -> None:
    if not (package_name or install_info):
        raise TypeError("Either 'package_name' or 'install_info' should be specified")
    install_info = install_info or InstallInfo(
        name=package_name,
        current_version="",
        new_version="",
        package=None,
    )
    message = "{} {}".format(  # pylint: disable=consider-using-f-string
        color_line("::", ColorsHighlight.yellow),
        translate("Ignoring package update {}").format(
            pretty_format_upgradeable(
                [install_info],
                template="{pkg_name} ({current_version} => {new_version})"
            ))
        if (install_info.current_version and install_info.new_version) else
        translate("Ignoring package {}").format(
            pretty_format_upgradeable(
                [install_info],
                template=(
                    "{pkg_name} {current_version}"
                    if install_info.current_version else
                    "{pkg_name} {new_version}"
                )
            ))
    )
    if ignored_from:
        message += f" {ignored_from}"
    print_stderr(message)


def _get_local_version(package_name: str) -> str:
    return PackageDB.get_local_dict()[package_name].version


def print_package_uptodate(package_name: str, package_source: PackageSource) -> None:
    print_warning(
        translate("{name} {version} {package_source} package is up to date - skipping").format(
            name=package_name,
            version=bold_line(_get_local_version(package_name)),
            package_source=package_source.name
        )
    )


def print_local_package_newer(package_name: str, aur_version: str) -> None:
    print_warning(
        translate(
            "{name} {version} local package is newer than in AUR ({aur_version}) - skipping"
        ).format(
            name=package_name,
            version=bold_line(_get_local_version(package_name)),
            aur_version=bold_line(aur_version),
        )
    )


def print_package_downgrading(package_name: str, downgrade_version: str) -> None:
    print_warning(
        translate("Downgrading AUR package {name} {version} to {downgrade_version}").format(
            name=bold_line(package_name),
            version=bold_line(_get_local_version(package_name)),
            downgrade_version=bold_line(downgrade_version)
        )
    )


def print_ignoring_outofdate_upgrade(package_info: InstallInfo) -> None:
    print_warning(
        translate("{name} {version} AUR package marked as 'outofdate' - skipping").format(
            name=package_info.name,
            version=bold_line(package_info.new_version),
        )
    )


# pylint:disable=too-many-locals,too-many-statements,too-many-branches
def print_package_search_results(
        repo_packages: Iterable[pyalpm.Package],
        aur_packages: Iterable[AURPackageInfo],
        local_pkgs_versions: dict[str, str],
        *,
        enumerated: bool = False,
) -> list[AnyPackage]:

    repos = [db.name for db in PackageDB.get_alpm_handle().get_syncdbs()]
    user_config = PikaurConfig()
    group_by_repo = user_config.ui.GroupByRepository.get_bool()

    def get_repo_sort_key(pkg: pyalpm.Package) -> tuple[int, str]:
        return (
            repos.index(pkg.db.name)
            if group_by_repo and pkg.db.name in repos
            else 999,
            pkg.name
        )

    # https://github.com/python/mypy/issues/11098
    type_aur_sort_key = tuple[float, float] | float  # type: ignore[misc]

    def get_aur_sort_key(pkg: AURPackageInfo) -> tuple[type_aur_sort_key, str]:
        user_aur_sort = user_config.ui.AurSearchSorting
        pkg_numvotes = pkg.numvotes if isinstance(pkg.numvotes, int) else 0
        pkg_popularity = pkg.popularity if isinstance(pkg.popularity, float) else 0.0

        if user_aur_sort == "pkgname":
            return (-1.0, pkg.name)
        if user_aur_sort == "popularity":
            return ((-pkg_popularity, -pkg_numvotes), pkg.name)
        if user_aur_sort == "numvotes":
            return ((-pkg_numvotes, -pkg_popularity), pkg.name)
        if user_aur_sort == "lastmodified":
            return (
                -pkg.lastmodified
                if isinstance(pkg.lastmodified, int)
                else 0,
                pkg.name)
        return (-(pkg_numvotes + 1) * (pkg_popularity + 1), pkg.name)

    args = parse_args()
    local_pkgs_names = local_pkgs_versions.keys()

    sorted_repo_pkgs: list[pyalpm.Package] = sorted(
        repo_packages,
        key=get_repo_sort_key
    )
    sorted_aur_pkgs: list[AURPackageInfo] = sorted(
        aur_packages,
        key=get_aur_sort_key
    )
    sorted_packages: list[AnyPackage] = [*sorted_repo_pkgs, *sorted_aur_pkgs]
    # mypy is always funny ^^ https://github.com/python/mypy/issues/5492#issuecomment-545992992

    enumerated_packages = list(enumerate(sorted_packages))
    if user_config.ui.ReverseSearchSorting.get_bool():
        enumerated_packages = list(reversed(enumerated_packages))

    for pkg_idx, package in enumerated_packages:
        # @TODO: return only packages for the current architecture
        idx = ""
        if enumerated:
            idx = bold_line(f"{pkg_idx+1}) ")

        pkg_name = package.name
        if args.quiet:
            print_stdout(f"{idx}{pkg_name}")
        else:

            repo = color_line("aur/", ColorsHighlight.red)
            if isinstance(package, pyalpm.Package):
                repo = pretty_format_repo_name(package.db.name)

            groups = ""
            if getattr(package, "groups", None):
                groups = color_line("({}) ".format(  # pylint: disable=consider-using-f-string
                    " ".join(package.groups)
                ), GROUP_COLOR)

            installed = ""
            if pkg_name in local_pkgs_names:
                if package.version != local_pkgs_versions[pkg_name]:
                    installed = color_line(
                        translate("[installed: {version}]").format(
                            version=local_pkgs_versions[pkg_name],
                        ) + " ",
                        ColorsHighlight.cyan
                    )
                else:
                    installed = color_line(
                        translate("[installed]") + " ",
                        ColorsHighlight.cyan
                    )

            rating = ""
            if (
                    isinstance(package, AURPackageInfo)
            ) and (
                package.numvotes is not None
            ) and (
                package.popularity is not None
            ):
                rating = color_line(
                    f"({package.numvotes}, {package.popularity:.2f})",
                    Colors.yellow
                )

            color_config = user_config.colors
            version_color = color_config.Version.get_int()
            version = package.version

            if isinstance(package, AURPackageInfo) and package.outofdate is not None:
                version_color = color_config.VersionDiffOld.get_int()
                version = "{} [{}: {}]".format(  # pylint: disable=consider-using-f-string
                    package.version,
                    translate("outofdate"),
                    datetime.fromtimestamp(package.outofdate).strftime("%Y/%m/%d")
                )

            last_updated = ""
            if user_config.ui.DisplayLastUpdated.get_bool():
                last_update_date = None

                if isinstance(package, pyalpm.Package):
                    last_update_date = package.builddate
                if isinstance(package, AURPackageInfo):
                    last_update_date = package.lastmodified

                last_updated = color_line(
                    " (last updated: {})".format(  # pylint: disable=consider-using-f-string
                        datetime.fromtimestamp(last_update_date).strftime("%Y/%m/%d")
                        if last_update_date is not None
                        else "unknown"
                    ),
                    ColorsHighlight.black
                )

            print_stdout("{}{}{} {} {}{}{}{}".format(  # pylint: disable=consider-using-f-string
                idx,
                repo,
                bold_line(pkg_name),
                color_line(version, version_color),
                groups,
                installed,
                rating,
                last_updated
            ))
            print_stdout(format_paragraph(f"{package.desc}"))
    return sorted_packages



#!/usr/bin/python3

"""Licensed under GPLv3, see https://www.gnu.org/licenses/"""

import atexit
import io
import os
import readline
import shutil
import signal
import sys
import traceback
from argparse import ArgumentError
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any, Callable

import pyalpm

from .args import parse_args
from .config import (
    _OLD_AUR_REPOS_CACHE_PATH,
    AUR_REPOS_CACHE_PATH,
    CACHE_ROOT,
    DATA_ROOT,
    PikaurConfig,
    get_config_path,
)
from .core import (
    DEFAULT_INPUT_ENCODING,
    check_runtime_deps,
    interactive_spawn,
    isolate_root_cmd,
    run_with_sudo_loop,
    running_as_root,
    spawn,
    sudo,
)
from .exceptions import SysExit
from .getpkgbuild_cli import cli_getpkgbuild
from .help_cli import cli_print_help
from .i18n import translate
from .info_cli import cli_info_packages
from .install_cli import InstallPackagesCLI
from .pikspect import PikspectSignalHandler, TTYRestore
from .pkg_cache_cli import cli_clean_packages_cache
from .pprint import create_debug_logger, print_error, print_stderr, print_warning
from .print_department import print_version
from .prompt import NotANumberInputError, get_multiple_numbers_input
from .search_cli import cli_search_packages, search_packages
from .updates import print_upgradeable
from .urllib import ProxyInitSocks5Error, init_proxy


def init_readline() -> None:
    # follow GNU readline config in prompts:
    system_inputrc_path = "/etc/inputrc"
    if os.path.exists(system_inputrc_path):
        readline.read_init_file(system_inputrc_path)
    user_inputrc_path = os.path.expanduser("~/.inputrc")
    if os.path.exists(user_inputrc_path):
        readline.read_init_file(user_inputrc_path)


init_readline()

_debug = create_debug_logger("main")


class OutputEncodingWrapper(AbstractContextManager[None]):

    original_stderr: int
    original_stdout: int

    def __enter__(self) -> None:
        for attr in ("stdout", "stderr"):
            _debug(f"Setting {attr} to {DEFAULT_INPUT_ENCODING}...", lock=False)
            real_stream = getattr(sys, attr)
            if real_stream.encoding == DEFAULT_INPUT_ENCODING:
                _debug("already set - nothing to do", lock=False)
                continue
            real_stream.flush()
            try:
                setattr(
                    self, f"original_{attr}",
                    real_stream
                )
                setattr(
                    sys, attr,
                    io.open(
                        real_stream.fileno(),
                        mode="w",
                        encoding=DEFAULT_INPUT_ENCODING,
                        closefd=False,
                    )
                )
            except io.UnsupportedOperation as exc:
                _debug(f"Can't set {attr} to {DEFAULT_INPUT_ENCODING}:\n{exc}", lock=False)

    def __exit__(
            self,
            exc_class: type | None, exc_instance: BaseException | None, exc_tb: TracebackType | None
    ) -> None:
        try:
            # @TODO: replace all SysExit-s to SystemExit-s eventually :3
            if exc_instance and exc_class and (exc_class not in (SysExit, SystemExit, )):
                # handling exception in context manager's __exit__ is not recommended
                # but otherwise stderr would be closed before exception is printed...
                if exc_tb:
                    print_stderr("".join(traceback.format_tb(exc_tb)), lock=False)
                print_stderr(f"{exc_class.__name__}: {exc_instance}", lock=False)
                sys.exit(121)
        finally:
            for attr in ("stdout", "stderr"):
                _debug(f"Restoring {attr}...", lock=False)
                stream = getattr(sys, attr)
                orig_stream = getattr(self, f"original_{attr}", None)
                if orig_stream in (None, stream, ):
                    _debug("nothing to do", lock=False)
                    continue
                stream.flush()
                setattr(
                    sys, attr,
                    orig_stream
                )
                _debug(f"{attr} restored", lock=False)
                stream.close()
                _debug(f"closed old {attr} stream", lock=False)


def cli_print_upgradeable() -> None:
    print_upgradeable()


def cli_install_packages() -> None:
    InstallPackagesCLI()


def cli_pkgbuild() -> None:
    cli_install_packages()


def cli_print_version() -> None:
    args = parse_args()
    proc = spawn(
        [PikaurConfig().misc.PacmanPath.get_str(), "--version", ],
    )
    if proc.stdout_text:
        pacman_version = proc.stdout_text.splitlines()[1].strip(" .-")
    else:
        pacman_version = "N/A"
    print_version(
        pacman_version=pacman_version, pyalpm_version=pyalpm.version(),
        quiet=args.quiet
    )


def cli_dynamic_select() -> None:  # pragma: no cover
    packages = search_packages(enumerated=True)
    if not packages:
        raise SysExit(1)

    while True:
        try:
            print_stderr(
                "\n" + translate(
                    "Please enter the number of the package(s) you want to install "
                    "and press [Enter] (default={}):"
                ).format(1)
            )
            answers = get_multiple_numbers_input("> ", list(range(1, len(packages) + 1))) or [1]
            print_stderr()
            selected_pkgs_idx = [idx - 1 for idx in answers]
            restart_prompt = False
            for idx in selected_pkgs_idx:
                if not 0 <= idx < len(packages):
                    print_error(translate("invalid value: {} is not between {} and {}").format(
                        idx + 1, 1, len(packages) + 1
                    ))
                    restart_prompt = True
            if restart_prompt:
                continue
            break
        except NotANumberInputError as exc:
            if exc.character.lower() == translate("n"):
                raise SysExit(128) from exc
            print_error(translate("invalid number: {}").format(exc.character))

    parse_args().positional = [packages[idx].name for idx in selected_pkgs_idx]
    cli_install_packages()


def cli_entry_point() -> None:  # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches

    try:
        init_proxy()
    except ProxyInitSocks5Error as exc:
        print_error("".join(exc.args))
        sys.exit(2)

    # operations are parsed in order what the less destructive (like info and query)
    # are being handled first, for cases when user by mistake
    # specified both operations, like `pikaur -QS smth`

    args = parse_args()
    pikaur_operation: Callable[[], None] | None = None
    require_sudo = False

    if args.help:
        pikaur_operation = cli_print_help

    elif args.version:
        pikaur_operation = cli_print_version

    elif args.query:
        if args.upgrades:
            pikaur_operation = cli_print_upgradeable

    elif args.files:
        require_sudo = bool(args.refresh)

    elif args.getpkgbuild:
        pikaur_operation = cli_getpkgbuild

    elif args.pkgbuild:
        require_sudo = True
        pikaur_operation = cli_pkgbuild

    elif args.sync:
        if args.search:
            pikaur_operation = cli_search_packages
        elif args.info:
            pikaur_operation = cli_info_packages
        elif args.clean:
            if not args.aur:
                require_sudo = True
            pikaur_operation = cli_clean_packages_cache
        elif args.groups or args.list:   # @TODO: implement -l/--list
            require_sudo = False
        else:
            require_sudo = True
            pikaur_operation = cli_install_packages

    elif not (args.database or args.remove or args.deptest or args.upgrade):
        if args.positional:
            pikaur_operation = cli_dynamic_select

    else:
        require_sudo = True

    if pikaur_operation:
        _debug(f"Pikaur operation found for {sys.argv=}: {pikaur_operation.__name__}")
        if require_sudo and args.dynamic_users and not running_as_root():
            # Restart pikaur with sudo to use systemd dynamic users
            restart_args = sys.argv[:]
            config_overridden = max(
                arg.startswith("--pikaur-config")
                for arg in restart_args
            )
            if not config_overridden:
                restart_args += ["--pikaur-config", get_config_path()]
            sys.exit(interactive_spawn(
                sudo(restart_args)
            ).returncode)
        else:
            if not require_sudo or running_as_root():
                # Just run the operation normally
                pikaur_operation()
            else:
                # Or use sudo loop if not running as root but need to have it later
                run_with_sudo_loop(pikaur_operation)
    else:
        # Just bypass all the args to pacman
        _debug(f"Pikaur operation not found for {sys.argv=}")
        _debug(args)
        pacman_args = [
            PikaurConfig().misc.PacmanPath.get_str(),
        ] + args.raw_without_pikaur_specific
        if require_sudo:
            pacman_args = sudo(pacman_args)
        sys.exit(
            interactive_spawn(pacman_args).returncode
        )


def migrate_old_aur_repos_dir() -> None:
    if not (
            os.path.exists(_OLD_AUR_REPOS_CACHE_PATH) and not os.path.exists(AUR_REPOS_CACHE_PATH)
    ):
        return
    if not os.path.exists(DATA_ROOT):
        os.makedirs(DATA_ROOT)
    shutil.move(_OLD_AUR_REPOS_CACHE_PATH, AUR_REPOS_CACHE_PATH)

    print_stderr()
    print_warning(
        translate(
            "AUR repos dir has been moved from '{old}' to '{new}'."
        ).format(
            old=_OLD_AUR_REPOS_CACHE_PATH,
            new=AUR_REPOS_CACHE_PATH
        )
    )
    print_stderr()


def create_dirs() -> None:
    if running_as_root():
        # Let systemd-run setup the directories and symlinks
        true_cmd = isolate_root_cmd(["true"])
        result = spawn(true_cmd)
        if result.returncode != 0:
            raise Exception(result)
        # Chown the private CacheDirectory to root to signal systemd that
        # it needs to recursively chown it to the correct user
        os.chown(os.path.realpath(CACHE_ROOT), 0, 0)
    if not os.path.exists(CACHE_ROOT):
        os.makedirs(CACHE_ROOT)
    migrate_old_aur_repos_dir()
    if not os.path.exists(AUR_REPOS_CACHE_PATH):
        os.makedirs(AUR_REPOS_CACHE_PATH)


def restore_tty() -> None:
    TTYRestore.restore()


def handle_sig_int(*_whatever: Any) -> None:  # pragma: no cover
    if signal_handler := PikspectSignalHandler.get():
        signal_handler(*_whatever)  # pylint: disable=not-callable
        return
    if parse_args().pikaur_debug:
        raise KeyboardInterrupt()
    print_stderr("\n\nCanceled by user (SIGINT)", lock=False)
    raise SysExit(125)


class EmptyWrapper:

    def __enter__(self) -> None:
        pass

    def __exit__(self, *_exc_details: Any) -> None:
        pass


def main(*, embed: bool = False) -> None:
    wrapper: type[AbstractContextManager[None]] = OutputEncodingWrapper
    if embed:
        wrapper = EmptyWrapper
    with wrapper():
        try:
            parse_args()
        except ArgumentError as exc:
            print_stderr(exc)
            sys.exit(22)
        check_runtime_deps()

        create_dirs()
        # initialize config to avoid race condition in threads:
        PikaurConfig.get_config()

        atexit.register(restore_tty)
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        signal.signal(signal.SIGINT, handle_sig_int)

        try:
            cli_entry_point()
        except BrokenPipeError:
            # @TODO: should it be 32?
            sys.exit(0)
        except SysExit as exc:
            sys.exit(exc.code)
        sys.exit(0)


if __name__ == "__main__":
    main()



import fcntl
import os
from typing import Any, TextIO

from .core import DEFAULT_INPUT_ENCODING
from .exceptions import SysExit
from .i18n import translate
from .pprint import print_error
from .prompt import ask_to_continue


class FileLock():

    locked = False
    lock_file: TextIO | None

    def __init__(self, lock_file_path: str) -> None:
        self.lock_file_path = lock_file_path

    def __enter__(self) -> None:
        self.lock_file = open(
            self.lock_file_path, "a", encoding=DEFAULT_INPUT_ENCODING
        )
        while True:
            try:
                fcntl.flock(self.lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                self.locked = True
                break
            except BlockingIOError as err:
                print_error(translate("Can't lock {lock_file}: {reason}").format(
                    lock_file=self.lock_file_path,
                    reason=err.strerror
                ))
                if not ask_to_continue(translate("Do you want to retry?")):
                    raise SysExit(128) from err

    def __exit__(self, *_exc_details: Any) -> None:
        if self.lock_file:
            if self.locked:
                fcntl.flock(self.lock_file, fcntl.LOCK_UN)
            if not self.lock_file.closed:
                self.lock_file.close()
        self.locked = False
        if os.path.exists(self.lock_file_path):
            os.remove(self.lock_file_path)

    def __del__(self) -> None:
        self.__exit__()
